"""Runway config models."""
# pylint: disable=no-self-argument
from __future__ import annotations

import locale
import logging
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generator,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
)

import yaml
from packaging.specifiers import InvalidSpecifier, SpecifierSet
from pydantic import Extra, Field, Protocol, root_validator, validator
from typing_extensions import Literal

from .. import utils
from ..base import ConfigProperty
from ..utils import RUNWAY_LOOKUP_STRING_ERROR, RUNWAY_LOOKUP_STRING_REGEX
from ._builtin_tests import (
    CfnLintRunwayTestArgs,
    CfnLintRunwayTestDefinitionModel,
    RunwayTestDefinitionModel,
    ScriptRunwayTestArgs,
    ScriptRunwayTestDefinitionModel,
    ValidRunwayTestTypeValues,
    YamlLintRunwayTestDefinitionModel,
)

if TYPE_CHECKING:
    from pydantic import BaseModel

    Model = TypeVar("Model", bound=BaseModel)

LOGGER = logging.getLogger(__name__)

RunwayEnvironmentsType = Dict[str, Union[bool, List[str], str]]
RunwayEnvironmentsUnresolvedType = Union[Dict[str, Union[bool, List[str], str]], str]
RunwayEnvVarsType = Dict[str, Union[List[str], str]]
RunwayEnvVarsUnresolvedType = Union[RunwayEnvVarsType, str]
RunwayModuleTypeTypeDef = Literal[
    "cdk", "cloudformation", "kubernetes", "serverless", "static", "terraform"
]

__all__ = [
    "CfnLintRunwayTestArgs",
    "CfnLintRunwayTestDefinitionModel",
    "RUNWAY_LOOKUP_STRING_ERROR",
    "RUNWAY_LOOKUP_STRING_REGEX",
    "RunwayAssumeRoleDefinitionModel",
    "RunwayConfigDefinitionModel",
    "RunwayDeploymentDefinitionModel",
    "RunwayDeploymentRegionDefinitionModel",
    "RunwayEnvironmentsType",
    "RunwayEnvironmentsUnresolvedType",
    "RunwayEnvVarsType",
    "RunwayEnvVarsUnresolvedType",
    "RunwayFutureDefinitionModel",
    "RunwayModuleDefinitionModel",
    "RunwayModuleTypeTypeDef",
    "RunwayTestDefinitionModel",
    "RunwayVariablesDefinitionModel",
    "RunwayVersionField",
    "ScriptRunwayTestArgs",
    "ScriptRunwayTestDefinitionModel",
    "ValidRunwayTestTypeValues",
    "YamlLintRunwayTestDefinitionModel",
]


class RunwayAssumeRoleDefinitionModel(ConfigProperty):
    """Model for a Runway assume role definition."""

    arn: Optional[str] = Field(
        default=None,
        title="IAM Role ARN",
        description="The ARN of the AWS IAM role to be assumed. (supports lookups)",
    )
    duration: Union[int, str] = Field(
        default=3600,
        description="The duration, in seconds, of the role session. (supports lookups)",
        ge=900,  # applies to int json schema only
        le=43_200,  # applies to int json schema only
        regex=RUNWAY_LOOKUP_STRING_REGEX,  # applies to str json schema only
    )
    post_deploy_env_revert: bool = Field(
        default=False,
        title="Post Deployment Environment Revert",
        description="Revert the credentials stored in environment variables to "
        "what they were prior to execution after the deployment finished processing. "
        "(supports lookups)",
    )
    session_name: str = Field(
        default="runway",
        description="An identifier for the assumed role session. (supports lookups)",
    )

    class Config(ConfigProperty.Config):
        """Model configuration."""

        extra = Extra.forbid
        schema_extra: Dict[str, Any] = {
            "description": "Used to defined a role to assume while Runway is "
            "processing each module.",
            "examples": [
                {"arn": "arn:aws:iam::123456789012:role/name"},
                {
                    "arn": "${var role_arn.${env DEPLOY_ENVIRONMENT}}",
                    "duration": 9001,
                    "post_deploy_env_revert": True,
                    "session_name": "runway-example",
                },
            ],
        }
        title = "Runway Deployment.assume_role Definition"

    @validator("arn")
    def _convert_arn_null_value(cls, v: Optional[str]) -> Optional[str]:
        """Convert a "nul" string into type(None)."""
        null_strings = ["null", "none", "undefined"]
        return None if isinstance(v, str) and v.lower() in null_strings else v

    @validator("duration", pre=True)
    def _validate_duration(cls, v: Union[int, str]) -> Union[int, str]:
        """Validate duration is within the range allowed by AWS."""
        if isinstance(v, str):
            return v
        if v < 900:
            raise ValueError("duration must be greater than or equal to 900")
        if v > 43_200:
            raise ValueError("duration must be less than or equal to 43,200")
        return v

    _validate_string_is_lookup = cast(
        "classmethod[Callable[..., Any]]",
        validator("duration", allow_reuse=True, pre=True)(
            utils.validate_string_is_lookup
        ),
    )


class RunwayDeploymentRegionDefinitionModel(ConfigProperty):
    """Model for a Runway deployment region definition."""

    parallel: Union[List[str], str] = Field(
        ...,
        title="Parallel Regions",
        description="An array of AWS Regions to process asynchronously. (supports lookups)",
    )

    class Config(ConfigProperty.Config):
        """Model configuration."""

        extra = Extra.forbid
        schema_extra: Dict[str, Any] = {
            "description": "Only supports 'parallel' field.",
            "examples": [
                {"parallel": ["us-east-1", "us-east-2"]},
                {"parallel": "${var regions.${env DEPLOY_ENVIRONMENT}}"},
            ],
        }
        title = "Runway Deployment.regions Definition"

    _validate_string_is_lookup = cast(
        "classmethod[Callable[..., Any]]",
        validator("parallel", allow_reuse=True, pre=True)(
            utils.validate_string_is_lookup
        ),
    )


class RunwayDeploymentDefinitionModel(ConfigProperty):
    """Model for a Runway deployment definition."""

    account_alias: Optional[str] = Field(
        default=None,
        description="Used to verify the currently assumed role or credentials. "
        "(supports lookups)",
        examples=["example-alias", "${var alias.${env DEPLOY_ENVIRONMENT}}"],
    )
    account_id: Optional[str] = Field(
        default=None,
        description="Used to verify the currently assumed role or credentials. "
        "(supports lookups)",
        examples=["123456789012", "${var id.${env DEPLOY_ENVIRONMENT}}"],
    )
    assume_role: Union[str, RunwayAssumeRoleDefinitionModel] = Field(
        default={},
        description="Assume a role when processing the deployment. (supports lookups)",
        examples=["arn:aws:iam::123456789012:role/name"]
        + cast(
            List[Any], RunwayAssumeRoleDefinitionModel.Config.schema_extra["examples"]
        ),
    )
    env_vars: RunwayEnvVarsUnresolvedType = Field(
        default={},
        title="Environment Variables",
        description="Additional variables to add to the environment when "
        "processing the deployment. (supports lookups)",
        examples=[
            "${var env_vars.${env DEPLOY_ENVIRONMENT}}",
            {
                "EXAMPLE_VARIABLE": "value",
                "KUBECONFIG": [".kube", "${env DEPLOY_ENVIRONMENT}", "config"],
            },
        ],
    )
    environments: RunwayEnvironmentsUnresolvedType = Field(
        default={},
        description="Explicitly enable/disable the deployment for a specific "
        "deploy environment, AWS Account ID, and AWS Region combination. "
        "Can also be set as a static boolean value. (supports lookups)",
        examples=[
            "${var envs.${env DEPLOY_ENVIRONMENT}}",
            {"dev": "123456789012", "prod": "us-east-1"},
            {"dev": True, "prod": False},
            {"dev": ["us-east-1"], "prod": ["us-west-2", "ca-central-1"]},
            {
                "dev": ["123456789012/us-east-1", "123456789012/us-west-2"],
                "prod": ["234567890123/us-east-1", "234567890123/us-west-2"],
            },
        ],
    )
    modules: List[RunwayModuleDefinitionModel] = Field(
        ..., description="An array of modules to process as part of a deployment."
    )
    module_options: Union[Dict[str, Any], str] = Field(
        default={},
        description="Options that are passed directly to the modules within this deployment. "
        "(supports lookups)",
        examples=[
            "${var sampleapp.options.${env DEPLOY_ENVIRONMENT}}",
            {"some_option": "value"},
        ],
    )
    name: str = Field(
        default="unnamed_deployment",
        description="The name of the deployment to be displayed in logs and the "
        "interactive selection menu.",
    )
    parallel_regions: Union[List[str], str] = Field(
        default=[],
        description="An array of AWS Regions to process asynchronously. (supports lookups)",
        examples=[
            ["us-east-1", "us-west-2"],
            "${var regions.${dev DEPLOY_ENVIRONMENT}}",
        ],
    )
    parameters: Union[Dict[str, Any], str] = Field(
        default={},
        description="Used to pass variable values to modules in place of an "
        "environment configuration file. (supports lookups)",
        examples=[
            {"namespace": "example-${env DEPLOY_ENVIRONMENT}"},
            "${var sampleapp.parameters.${env DEPLOY_ENVIRONMENT}}",
        ],
    )
    regions: Union[List[str], str] = Field(
        default=[],
        description="An array of AWS Regions to process this deployment in. (supports lookups)",
        examples=[
            ["us-east-1", "us-west-2"],
            "${var regions.${dev DEPLOY_ENVIRONMENT}}",
        ]
        + RunwayDeploymentRegionDefinitionModel.Config.schema_extra["examples"],
    )

    class Config(ConfigProperty.Config):
        """Model configuration."""

        extra = Extra.forbid
        title = "Runway Deployment Definition"

        @staticmethod
        def schema_extra(schema: Dict[str, Any]) -> None:  # type: ignore
            """Processess the schema after it has been generated.

            Schema is modified in place. Return value is ignored.

            https://pydantic-docs.helpmanual.io/usage/schema/#schema-customization

            """
            schema[
                "description"
            ] = "A collection of modules, regions, and other configurations to deploy."
            # modify schema to allow simple string or mapping definition for a module
            module_ref = schema["properties"]["modules"]["items"].pop("$ref")
            schema["properties"]["modules"]["items"]["anyOf"] = [
                {"$ref": module_ref},
                {"type": "string"},
            ]

    @root_validator(pre=True)
    def _convert_simple_module(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Convert simple modules to dicts."""
        modules = values.get("modules", [])
        result: List[Dict[str, Any]] = []
        for module in modules:
            if isinstance(module, str):
                result.append({"path": module})
            else:
                result.append(module)
        values["modules"] = result
        return values

    @root_validator(pre=True)
    def _validate_regions(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Validate & simplify regions."""
        raw_regions = values.get("regions", [])
        parallel_regions = values.get("parallel_regions", [])
        if all(isinstance(i, str) for i in [raw_regions, parallel_regions]):
            raise ValueError(
                "unable to validate parallel_regions/regions - both are defined as strings"
            )
        if any(isinstance(i, str) for i in [raw_regions, parallel_regions]):
            return values  # one is a lookup so skip the remainder of the checks
        if isinstance(raw_regions, list):
            regions = raw_regions
        else:
            regions = RunwayDeploymentRegionDefinitionModel.parse_obj(raw_regions)

        if regions and parallel_regions:
            raise ValueError("only one of parallel_regions or regions can be defined")
        if not regions and not parallel_regions:
            raise ValueError("either parallel_regions or regions must be defined")

        if isinstance(regions, RunwayDeploymentRegionDefinitionModel):
            values["regions"] = []
            values["parallel_regions"] = regions.parallel
        return values

    _validate_string_is_lookup = cast(
        "classmethod[Callable[..., Any]]",
        validator(
            "env_vars",
            "environments",
            "module_options",
            "parallel_regions",
            "parameters",
            "regions",
            allow_reuse=True,
            pre=True,
        )(utils.validate_string_is_lookup),
    )


class RunwayFutureDefinitionModel(ConfigProperty):
    """Model for the Runway future definition."""

    class Config(ConfigProperty.Config):
        """Model configuration."""

        extra = Extra.forbid
        schema_extra = {
            "description": "Enable features/behaviors that will be become standard "
            "ahead of their official release."
        }
        title = "Runway Future Definition"


class RunwayModuleDefinitionModel(ConfigProperty):
    """Model for a Runway module definition."""

    class_path: Optional[str] = Field(
        default=None,
        description="Import path to a custom Runway module class. (supports lookups)",
    )
    env_vars: RunwayEnvVarsUnresolvedType = Field(
        default={},
        title="Environment Variables",
        description="Additional variables to add to the environment when "
        "processing the deployment. (supports lookups)",
        examples=[
            "${var env_vars.${env DEPLOY_ENVIRONMENT}}",
            {
                "EXAMPLE_VARIABLE": "value",
                "KUBECONFIG": [".kube", "${env DEPLOY_ENVIRONMENT}", "config"],
            },
        ],
    )
    environments: RunwayEnvironmentsUnresolvedType = Field(
        default={},
        description="Explicitly enable/disable the deployment for a specific "
        "deploy environment, AWS Account ID, and AWS Region combination. "
        "Can also be set as a static boolean value. (supports lookups)",
        examples=[
            "${var envs.${env DEPLOY_ENVIRONMENT}}",
            {"dev": "123456789012", "prod": "us-east-1"},
            {"dev": True, "prod": False},
            {"dev": ["us-east-1"], "prod": ["us-west-2", "ca-central-1"]},
            {
                "dev": ["123456789012/us-east-1", "123456789012/us-west-2"],
                "prod": ["234567890123/us-east-1", "234567890123/us-west-2"],
            },
        ],
    )
    name: str = Field(
        default="undefined",
        description="The name of the module to be displayed in logs and the "
        "interactive selection menu.",
    )
    options: Union[Dict[str, Any], str] = Field(
        default={}, description="Module type specific options. (supports lookups)"
    )
    parameters: Union[Dict[str, Any], str] = Field(
        default={},
        description="Used to pass variable values to modules in place of an "
        "environment configuration file. (supports lookups)",
        examples=[
            {"namespace": "example-${env DEPLOY_ENVIRONMENT}"},
            "${var sampleapp.parameters.${env DEPLOY_ENVIRONMENT}}",
        ],
    )
    path: Optional[Union[str, Path]] = Field(
        default=None,
        description="Directory (relative to the Runway config file) containing IaC. "
        "(supports lookups)",
        examples=["./", "sampleapp-${env DEPLOY_ENVIRONMENT}.cfn", "sampleapp.sls"],
    )
    tags: List[str] = Field(
        default=[],
        description="Array of values to categorize the module which can be used "
        "with the CLI to quickly select a group of modules. "
        "This field is only used by the `--tag` CLI option.",
        examples=[["type:network", "app:sampleapp"]],
    )
    type: Optional[RunwayModuleTypeTypeDef] = None
    # needs to be last
    parallel: List[RunwayModuleDefinitionModel] = Field(
        default=[],
        description="Array of module definitions that can be executed asynchronously. "
        "Incompatible with class_path, path, and type.",
        examples=[[{"path": "sampleapp-01.cfn"}, {"path": "sampleapp-02.cfn"}]],
    )

    class Config(ConfigProperty.Config):
        """Model configuration."""

        extra = Extra.forbid
        schema_extra = {
            "description": "Defines a directory containing IaC, "
            "the parameters to pass in during execution, "
            "and any applicable options for the module type.",
        }
        title = "Runway Module Definition"
        use_enum_values = True

    @root_validator(pre=True)
    def _validate_name(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Validate module name."""
        if "name" in values:
            return values
        if "parallel" in values:
            values["name"] = "parallel_parent"
            return values
        if "path" in values:
            values["name"] = Path(values["path"]).resolve().name
            return values
        return values

    @root_validator(pre=True)
    def _validate_path(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Validate path and sets a default value if needed."""
        if not values.get("path") and not values.get("parallel"):
            values["path"] = Path.cwd()
        return values

    @validator("parallel", pre=True)
    def _validate_parallel(
        cls, v: List[Union[Dict[str, Any], str]], values: Dict[str, Any]
    ) -> List[Dict[str, Any]]:
        """Validate parallel."""
        if v and values.get("path"):
            raise ValueError("only one of parallel or path can be defined")
        result: List[Dict[str, Any]] = []
        for mod in v:
            if isinstance(mod, str):
                result.append({"path": mod})
            else:
                result.append(mod)
        return result

    # TODO add regex to schema
    _validate_string_is_lookup = cast(
        "classmethod[Callable[..., Any]]",
        validator(
            "env_vars",
            "environments",
            "options",
            "parameters",
            allow_reuse=True,
            pre=True,
        )(utils.validate_string_is_lookup),
    )


# https://pydantic-docs.helpmanual.io/usage/postponed_annotations/#self-referencing-models
RunwayModuleDefinitionModel.update_forward_refs()


class RunwayVariablesDefinitionModel(ConfigProperty):
    """Model for a Runway variable definition."""

    file_path: Optional[Path] = Field(
        default=None,
        title="Variables File Path",
        description="Explicit path to a variables file that will be loaded and "
        "merged with the variables defined here.",
    )
    sys_path: Path = Field(
        default="./",
        description="Directory to use as the root of a relative 'file_path'. "
        "If not provided, the current working directory is used.",
    )

    class Config(ConfigProperty.Config):
        """Model configuration."""

        extra = Extra.allow
        schema_extra = {
            "description": "A variable definitions for the Runway config file. "
            "This is used to resolve the 'var' lookup.",
        }
        title = "Runway Variables Definition"

    _convert_null_values = cast(
        "classmethod[Callable[..., Any]]",
        validator("*", allow_reuse=True)(utils.convert_null_values),
    )


class RunwayVersionField(SpecifierSet):
    """Extends packaging.specifiers.SpecifierSet for use with pydantic."""

    @classmethod
    def __get_validators__(cls) -> Generator[Callable[..., Any], None, None]:
        """Yield one of more validators with will be called to validate the input.

        Each validator will receive, as input, the value returned from the previous validator.

        """
        yield cls._convert_value

    @classmethod
    def __modify_schema__(cls, field_schema: Dict[str, Any]) -> None:
        """Mutate the field schema in place.

        This is only called when output JSON schema from a model.

        """
        field_schema.update(type="string")  # cov: ignore

    @classmethod
    def _convert_value(cls, v: Union[str, SpecifierSet]) -> RunwayVersionField:
        """Convert runway_version string into SpecifierSet with some value handling.

        Args:
            v: The value to be converted/validated.

        Raises:
            ValueError: The provided value is not a valid version specifier set.

        """
        if isinstance(v, (cls, SpecifierSet)):
            return RunwayVersionField(str(v), prereleases=True)
        try:
            return RunwayVersionField(v, prereleases=True)
        except InvalidSpecifier:
            if any(v.startswith(i) for i in ["!", "~", "<", ">", "="]):
                raise ValueError(f"{v} is not a valid version specifier set") from None
            LOGGER.debug(
                "runway_version is not a valid version specifier; trying as an exact version",
                exc_info=True,
            )
            return RunwayVersionField("==" + v, prereleases=True)


class RunwayConfigDefinitionModel(ConfigProperty):
    """Runway configuration definition model."""

    deployments: List[RunwayDeploymentDefinitionModel] = Field(
        default=[], description="Array of Runway deployments definitions."
    )
    future: RunwayFutureDefinitionModel = RunwayFutureDefinitionModel()
    ignore_git_branch: bool = Field(
        default=False,
        description="Optionally exclude the git branch name when determining the "
        "current deploy environment.",
    )
    runway_version: Optional[RunwayVersionField] = Field(
        default=">1.10",
        description="Define the versions of Runway that can be used with this "
        "configuration file.",
        examples=['"<2.0.0"', '"==1.14.0"', '">=1.14.0,<2.0.0"'],
    )
    tests: List[RunwayTestDefinitionModel] = Field(
        default=[],
        description="Array of Runway test definitions that are executed with the 'test' command.",
    )
    variables: RunwayVariablesDefinitionModel = RunwayVariablesDefinitionModel()

    class Config(ConfigProperty.Config):
        """Model configuration."""

        extra = Extra.forbid
        schema_extra = {
            "description": "Configuration file for use with Runway.",
        }
        title = "Runway Configuration File"
        validate_all = True
        validate_assignment = True

    @root_validator(pre=True)
    def _add_deployment_names(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Add names to deployments that are missing them."""
        deployments = values.get("deployments", [])
        for i, deployment in enumerate(deployments):
            if not deployment.get("name"):
                deployment["name"] = f"deployment_{i + 1}"
        values["deployments"] = deployments
        return values

    @classmethod
    def parse_file(
        cls: Type[Model],
        path: Union[str, Path],
        *,
        content_type: Optional[str] = None,
        encoding: str = "utf8",
        proto: Optional[Protocol] = None,
        allow_pickle: bool = False,
    ) -> Model:
        """Parse a file."""
        return cast(
            "Model",
            cls.parse_raw(
                Path(path).read_text(
                    encoding=locale.getpreferredencoding(do_setlocale=False)
                ),
                content_type=content_type,  # type: ignore
                encoding=encoding,
                proto=proto,  # type: ignore
                allow_pickle=allow_pickle,
            ),
        )

    @classmethod
    def parse_raw(
        cls: Type[Model],
        b: Union[bytes, str],
        *,
        content_type: Optional[str] = None,  # pylint: disable=unused-argument
        encoding: str = "utf8",  # pylint: disable=unused-argument
        proto: Optional[Protocol] = None,  # pylint: disable=unused-argument
        allow_pickle: bool = False,  # pylint: disable=unused-argument
    ) -> Model:
        """Parse raw data."""
        return cast("Model", cls.parse_obj(yaml.safe_load(b)))


# https://pydantic-docs.helpmanual.io/usage/postponed_annotations/#self-referencing-models
RunwayDeploymentDefinitionModel.update_forward_refs()



"""cfn-list test runner."""
from __future__ import annotations

import locale
import logging
import runpy
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Union

import yaml

from ..._logging import PrefixAdaptor
from ...utils import argv
from .base import TestHandler

if TYPE_CHECKING:
    from ...config.components.runway.base import ConfigProperty

TYPE_NAME = "cfn-lint"
LOGGER = logging.getLogger(__name__)


class CfnLintHandler(TestHandler):
    """Lints CFN."""

    @classmethod
    def handle(cls, name: str, args: Union[ConfigProperty, Dict[str, Any]]) -> None:
        """Perform the actual test.

        Relies on .cfnlintrc file to be located beside the Runway config file.

        """
        logger = PrefixAdaptor(name, LOGGER)
        cfnlintrc = Path("./.cfnlintrc")

        if not cfnlintrc.is_file():
            logger.error("file must exist to use this test: %s", cfnlintrc)
            sys.exit(1)

        # prevent duplicate log messages by not passing to the root logger
        logging.getLogger("cfnlint").propagate = False
        try:
            with argv(*["cfn-lint"] + args.get("cli_args", [])):
                runpy.run_module("cfnlint", run_name="__main__")
        except SystemExit as err:  # this call will always result in SystemExit
            if err.code != 0:  # ignore zero exit codes but re-raise for non-zero
                if not (
                    yaml.safe_load(
                        cfnlintrc.read_text(
                            encoding=locale.getpreferredencoding(do_setlocale=False)
                        )
                    )
                    or {}
                ).get("templates"):
                    logger.warning(
                        'cfnlintrc is missing a "templates" '
                        "section which is required by cfn-lint"
                    )
                raise



"""Replicated Lambda Function cleanup warning."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, List

from ..base import HookArgsBaseModel

if TYPE_CHECKING:
    from mypy_boto3_cloudformation.type_defs import OutputTypeDef

    from ....context import CfnginContext

LOGGER = logging.getLogger(__name__)
REPLICATED_FUNCTION_OUTPUTS = [
    "LambdaCheckAuthArn",
    "LambdaHttpHeadersArn",
    "LambdaParseAuthArn",
    "LambdaRefreshAuthArn",
    "LambdaSignOutArn",
    "LambdaCFDirectoryIndexRewriteArn",
]
STACK_STATUSES_TO_IGNORE = [
    "ROLLBACK_IN_PROGRESS",
    "ROLLBACK_FAILED",
    "ROLLBACK_COMPLETE",
    "DELETE_IN_PROGRESS",
    "DELETE_FAILED",
    "DELETE_COMPLETE",
    "IMPORT_ROLLBACK_IN_PROGRESS",
    "IMPORT_ROLLBACK_FAILED",
    "IMPORT_ROLLBACK_COMPLETE",
]


class HookArgs(HookArgsBaseModel):
    """Hook arguments."""

    stack_relative_name: str
    """Name of the CloudFormation Stack as defined in the config file (no namespace)."""


def get_replicated_function_names(outputs: List[OutputTypeDef]) -> List[str]:
    """Extract replicated function names from CFN outputs."""
    function_names: List[str] = []
    for i in REPLICATED_FUNCTION_OUTPUTS:
        function_arn = next(
            (
                output.get("OutputValue")
                for output in outputs
                if output.get("OutputKey") == i
            ),
            None,
        )
        if function_arn:
            function_names.append(function_arn.split(":")[-1])
    return function_names


def warn(context: CfnginContext, *__args: Any, **kwargs: Any) -> bool:
    """Notify the user of Lambda functions to delete.

    Arguments parsed by :class:`~runway.cfngin.hooks.staticsite.cleanup.HookArgs`.

    Args:
        context: The context instance.

    """
    args = HookArgs.parse_obj(kwargs)
    cfn_client = context.get_session().client("cloudformation")
    try:
        describe_response = cfn_client.describe_stacks(
            StackName=context.namespace
            + context.namespace_delimiter
            + args.stack_relative_name
        )
        stack = next(
            x
            for x in describe_response.get("Stacks", [])
            if (
                x.get("StackStatus")
                and x.get("StackStatus") not in STACK_STATUSES_TO_IGNORE
            )
        )
        functions = get_replicated_function_names(stack["Outputs"])
        if functions:
            cmd = (
                "aws lambda delete-function --function-name $x "
                f"--region {context.env.aws_region}"
            )
            LOGGER.warning(
                "About to delete the Static Site stack that contains "
                "replicated Lambda functions. These functions cannot "
                "be deleted until AWS automatically deletes their "
                "replicas. After some time has passed (a day is "
                "typically sufficient), they can be manually deleted. "
                "E.g.:"
            )
            LOGGER.warning("On macOS/Linux:")
            LOGGER.warning("for x in %s; do %s; done", (" ").join(functions), cmd)
            LOGGER.warning("On Windows:")
            LOGGER.warning('Foreach ($x in "%s") { %s }', ('","').join(functions), cmd)
    except Exception:  # pylint: disable=broad-except
        # There's no harm in continuing on in the event of an error
        # Orphaned functions have no cost
        pass
    return True



"""``runway schema`` command group."""
from typing import Any

import click

from ... import options
from ._cfngin import cfngin
from ._runway import runway

COMMANDS = [cfngin, runway]


@click.group("schema", short_help="JSON schema")
@options.debug
@options.no_color
@options.verbose
def schema(**_: Any) -> None:
    """Output JSON schema for Runway or CFNgin configuration files."""


for cmd in COMMANDS:  # register commands
    schema.add_command(cmd)



"""CFNgin lookups."""
# export resolve_lookups at this level
from .registry import register_lookup_handler, unregister_lookup_handler

__all__ = ["register_lookup_handler", "unregister_lookup_handler"]



"""Handler for fetching outputs from fully qualified stacks."""
# pyright: reportIncompatibleMethodOverride=none
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from typing_extensions import Final, Literal

from ....lookups.handlers.base import LookupHandler
from .output import deconstruct

if TYPE_CHECKING:
    from ...providers.aws.default import Provider

LOGGER = logging.getLogger(__name__)

XREF_PRESISTENT_STATE = {"has_warned": False}


class XrefLookup(LookupHandler):
    """Xref lookup."""

    DEPRECATION_MSG = "xref Lookup has been deprecated; use the cfn lookup instead"
    TYPE_NAME: Final[Literal["xref"]] = "xref"
    """Name that the Lookup is registered as."""

    @classmethod
    def handle(  # pylint: disable=arguments-differ,arguments-renamed
        cls, value: str, provider: Provider, **_: Any
    ) -> str:
        """Fetch an output from the designated, fully qualified stack.

        The `output` handler supports fetching outputs from stacks created
        within a single config file. Sometimes it's useful to fetch outputs
        from stacks created outside of the current config file. `xref`
        supports this by **not** using the
        :class:`runway.context.CfnginContext` to expand the fqn of the stack.

        Args:
            value: Parameter(s) given to this lookup. ``<stack_name>::<output_name>``
            provider: Provider instance.

        Returns:
            Output from the specified stack.

        Example:
            ::

                conf_value: ${xref fully-qualified-stack-name::SomeOutputName}

        """
        decon = deconstruct(value)
        stack_fqn = decon.stack_name
        return provider.get_output(stack_fqn, decon.output_name)



"""CFNgin entrypoint."""
from __future__ import annotations

import logging
import os
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, cast

from yaml.constructor import ConstructorError

from .._logging import PrefixAdaptor
from ..compat import cached_property
from ..config import CfnginConfig
from ..context import CfnginContext
from ..utils import MutableMap, SafeHaven
from .actions import deploy, destroy, diff, init
from .environment import parse_environment
from .providers.aws.default import ProviderBuilder

if TYPE_CHECKING:
    from .._logging import RunwayLogger
    from ..context import RunwayContext

# explicitly name logger so its not redundant
LOGGER = cast("RunwayLogger", logging.getLogger("runway.cfngin"))


class CFNgin:
    """Control CFNgin.

    Attributes:
        concurrency: Max number of CFNgin stacks that can be deployed concurrently.
            If the value is ``0``, will be constrained based on the underlying graph.
        interactive: Whether or not to prompt the user before taking action.
        parameters: Combination of the parameters provided when initalizing the
            class and any environment files that are found.
        recreate_failed: Destroy and re-create stacks that are stuck in
            a failed state from an initial deployment when updating.
        region: The AWS region where CFNgin is currently being executed.
        sys_path: Working directory.
        tail: Whether or not to display all CloudFormation events in the terminal.

    """

    concurrency: int
    interactive: bool
    parameters: MutableMap
    recreate_failed: bool
    region: str
    sys_path: Path
    tail: bool

    def __init__(
        self,
        ctx: RunwayContext,
        parameters: Optional[Dict[str, Any]] = None,
        sys_path: Optional[Path] = None,
    ) -> None:
        """Instantiate class.

        Args:
            ctx: Runway context object.
            parameters: Parameters from Runway.
            sys_path: Working directory.

        """
        self.__ctx = ctx
        self._env_file_name = None
        self.concurrency = ctx.env.max_concurrent_cfngin_stacks
        self.interactive = ctx.is_interactive
        self.parameters = MutableMap()
        self.recreate_failed = ctx.is_noninteractive
        self.region = ctx.env.aws_region
        self.sys_path = sys_path or Path.cwd()
        self.tail = bool(ctx.env.debug or ctx.env.verbose)

        self.parameters.update(self.env_file)

        if parameters:
            LOGGER.debug("adding Runway parameters to CFNgin parameters")
            self.parameters.update(parameters)

        self._inject_common_parameters()

    @cached_property
    def env_file(self) -> MutableMap:
        """Contents of a CFNgin environment file."""
        result: Dict[str, Any] = {}
        supported_names = [
            f"{self.__ctx.env.name}.env",
            f"{self.__ctx.env.name}-{self.region}.env",
        ]
        for _, file_name in enumerate(supported_names):
            file_path = os.path.join(self.sys_path, file_name)
            if os.path.isfile(file_path):
                LOGGER.info("found environment file: %s", file_path)
                self._env_file_name = file_path
                with open(file_path, "r", encoding="utf-8") as file_:
                    result.update(parse_environment(file_.read()))
        return MutableMap(**result)

    def deploy(self, force: bool = False, sys_path: Optional[Path] = None) -> None:
        """Run the CFNgin deploy action.

        Args:
            force: Explicitly enable the action even if an environment
                file is not found.
            sys_path: Explicitly define a path to work in.
                If not provided, ``self.sys_path`` is used.

        """
        if self.should_skip(force):
            return
        sys_path = sys_path or self.sys_path
        config_file_paths = self.find_config_files(sys_path=sys_path)

        with SafeHaven(
            environ=self.__ctx.env.vars, sys_modules_exclude=["awacs", "troposphere"]
        ):
            for config_path in config_file_paths:
                logger = PrefixAdaptor(os.path.basename(config_path), LOGGER)
                logger.notice("deploy (in progress)")
                with SafeHaven(sys_modules_exclude=["awacs", "troposphere"]):
                    ctx = self.load(config_path)
                    action = deploy.Action(
                        context=ctx,
                        provider_builder=self._get_provider_builder(
                            ctx.config.service_role
                        ),
                    )
                    action.execute(concurrency=self.concurrency, tail=self.tail)
                logger.success("deploy (complete)")

    def destroy(self, force: bool = False, sys_path: Optional[Path] = None) -> None:
        """Run the CFNgin destroy action.

        Args:
            force: Explicitly enable the action even if an environment
                file is not found.
            sys_path: Explicitly define a path to work in.
                If not provided, ``self.sys_path`` is used.

        """
        if self.should_skip(force):
            return
        sys_path = sys_path or self.sys_path
        config_file_paths = self.find_config_files(sys_path=sys_path)
        # destroy should run in reverse to handle dependencies
        config_file_paths.reverse()

        with SafeHaven(environ=self.__ctx.env.vars):
            for config_path in config_file_paths:
                logger = PrefixAdaptor(config_path.name, LOGGER)
                logger.notice("destroy (in progress)")
                with SafeHaven():
                    ctx = self.load(config_path)
                    action = destroy.Action(
                        context=ctx,
                        provider_builder=self._get_provider_builder(
                            ctx.config.service_role
                        ),
                    )
                    action.execute(
                        concurrency=self.concurrency, force=True, tail=self.tail
                    )
                logger.success("destroy (complete)")

    def init(self, force: bool = False, sys_path: Optional[Path] = None) -> None:
        """Initialize environment."""
        if self.should_skip(force):
            return
        sys_path = sys_path or self.sys_path
        config_file_paths = self.find_config_files(sys_path=sys_path)

        with SafeHaven(
            environ=self.__ctx.env.vars, sys_modules_exclude=["awacs", "troposphere"]
        ):
            for config_path in config_file_paths:
                logger = PrefixAdaptor(os.path.basename(config_path), LOGGER)
                logger.notice("init (in progress)")
                with SafeHaven(sys_modules_exclude=["awacs", "troposphere"]):
                    ctx = self.load(config_path)
                    action = init.Action(
                        context=ctx,
                        provider_builder=self._get_provider_builder(
                            ctx.config.service_role
                        ),
                    )
                    action.execute(concurrency=self.concurrency, tail=self.tail)
                logger.success("init (complete)")

    def load(self, config_path: Path) -> CfnginContext:
        """Load a CFNgin config into a context object.

        Args:
            config_path: Valid path to a CFNgin config file.

        """
        LOGGER.debug("loading CFNgin config: %s", config_path.name)
        try:
            config = self._get_config(config_path)
            config.load()
            return self._get_context(config, config_path)
        except ConstructorError as err:
            if str(err.problem).startswith(  # could be NoneType
                "could not determine a constructor for the tag '!"
            ):
                LOGGER.error(
                    '"%s" is located in the module\'s root directory '
                    "and appears to be a CloudFormation template; "
                    "please move CloudFormation templates to a subdirectory",
                    config_path,
                )
                sys.exit(1)
            raise

    def plan(self, force: bool = False, sys_path: Optional[Path] = None):
        """Run the CFNgin plan action.

        Args:
            force: Explicitly enable the action even if an environment
                file is not found.
            sys_path: Explicitly define a path to work in.
                If not provided, ``self.sys_path`` is used.

        """
        if self.should_skip(force):
            return
        sys_path = sys_path or self.sys_path
        config_file_paths = self.find_config_files(sys_path=sys_path)
        with SafeHaven(environ=self.__ctx.env.vars):
            for config_path in config_file_paths:
                logger = PrefixAdaptor(config_path.name, LOGGER)
                logger.notice("plan (in progress)")
                with SafeHaven():
                    ctx = self.load(config_path)
                    action = diff.Action(
                        context=ctx,
                        provider_builder=self._get_provider_builder(
                            ctx.config.service_role
                        ),
                    )
                    action.execute()
                logger.success("plan (complete)")

    def should_skip(self, force: bool = False) -> bool:
        """Determine if action should be taken or not.

        Args:
            force: If ``True``, will always return ``False`` meaning
                the action should not be skipped.

        """
        if force or self.env_file:
            return False
        LOGGER.info("skipped; no parameters and environment file not found")
        return True

    def _get_config(self, file_path: Path) -> CfnginConfig:
        """Initialize a CFNgin config object from a file.

        Args:
            file_path: Path to the config file to load.
            validate: Validate the loaded config.

        """
        return CfnginConfig.parse_file(
            file_path=file_path,
            parameters=self.parameters,
            work_dir=self.__ctx.work_dir,
        )

    def _get_context(self, config: CfnginConfig, config_path: Path) -> CfnginContext:
        """Initialize a CFNgin context object.

        Args:
            config: CFNgin config object.
            config_path: Path to the config file that was provided.

        """
        return CfnginContext(
            config_path=config_path,
            config=config,
            deploy_environment=self.__ctx.env.copy(),
            force_stacks=[],  # placeholder
            parameters=self.parameters,
            stack_names=[],  # placeholder
            work_dir=self.__ctx.work_dir,
        )

    def _get_provider_builder(
        self, service_role: Optional[str] = None
    ) -> ProviderBuilder:
        """Initialize provider builder.

        Args:
            service_role: CloudFormation service role.

        """
        if self.interactive:
            LOGGER.verbose("using interactive AWS provider mode")
        else:
            LOGGER.verbose("using default AWS provider mode")
        return ProviderBuilder(
            interactive=self.interactive,
            recreate_failed=self.recreate_failed,
            region=self.region,
            service_role=service_role,
        )

    def _inject_common_parameters(self) -> None:
        """Add common parameters if they don't already exist.

        Adding these commonly used parameters will remove the need to add
        lookup support (mainly for environment variable lookups) in places
        such as ``cfngin_bucket``.

        Injected Parameters
        ~~~~~~~~~~~~~~~~~~~

        **environment (str)**
            Taken from the ``DEPLOY_ENVIRONMENT`` environment variable. This
            will the be current Runway environment being processed.

        **region (str)**
            Taken from the ``AWS_REGION`` environment variable. This will be
            the current region being deployed to.

        """
        if not self.parameters.get("environment"):
            self.parameters["environment"] = self.__ctx.env.name
        if not self.parameters.get("region"):
            self.parameters["region"] = self.region

    @classmethod
    def find_config_files(
        cls, exclude: Optional[List[str]] = None, sys_path: Optional[Path] = None
    ) -> List[Path]:
        """Find CFNgin config files.

        Args:
            exclude: List of file names to exclude. This list is appended to
                the global exclude list.
            sys_path: Explicitly define a path to search for config files.

        Returns:
            Paths to config files that were found.

        """
        return CfnginConfig.find_config_file(sys_path, exclude=exclude)



"""Runway command import aggregation."""
from ._deploy import deploy
from ._destroy import destroy
from ._dismantle import dismantle
from ._docs import docs
from ._envvars import envvars
from ._gen_sample import gen_sample
from ._init import init
from ._kbenv import kbenv
from ._new import new
from ._plan import plan
from ._preflight import preflight
from ._run_python import run_python
from ._schema import schema
from ._takeoff import takeoff
from ._taxi import taxi
from ._test import test
from ._tfenv import tfenv
from ._whichenv import whichenv

__all__ = [
    "deploy",
    "destroy",
    "dismantle",
    "docs",
    "envvars",
    "gen_sample",
    "init",
    "kbenv",
    "new",
    "plan",
    "preflight",
    "run_python",
    "schema",
    "takeoff",
    "taxi",
    "test",
    "tfenv",
    "whichenv",
]



"""Runway config deployment definition."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Tuple, Union, overload

from ....exceptions import UnresolvedVariable
from ....variables import Variable
from ...models.runway import (
    RunwayDeploymentDefinitionModel,
    RunwayModuleDefinitionModel,
)
from ._module_def import RunwayModuleDefinition
from .base import ConfigComponentDefinition

if TYPE_CHECKING:
    from ...models.base import ConfigProperty
    from ...models.runway import (
        RunwayAssumeRoleDefinitionModel,
        RunwayEnvironmentsType,
        RunwayEnvVarsType,
    )

LOGGER = logging.getLogger(__name__.replace("._", "."))


class RunwayDeploymentDefinition(ConfigComponentDefinition):
    """Runway deployment definition."""

    account_alias: Optional[str]
    account_id: Optional[str]
    assume_role: RunwayAssumeRoleDefinitionModel
    environments: RunwayEnvironmentsType
    env_vars: RunwayEnvVarsType
    module_options: Dict[str, Any]
    name: str
    parallel_regions: List[str]
    parameters: Dict[str, Any]
    regions: List[str]

    _data: RunwayDeploymentDefinitionModel
    _pre_process_vars: Tuple[str, ...] = (
        "account_alias",
        "account_id",
        "assume_role",
        "env_vars",
        "regions",
    )
    _supports_vars: Tuple[str, ...] = (
        "account_alias",
        "account_id",
        "assume_role",
        "env_vars",
        "environments",
        "module_options",
        "regions",
        "parallel_regions",
        "parameters",
    )

    def __init__(self, data: RunwayDeploymentDefinitionModel) -> None:
        """Instantiate class."""
        super().__init__(data)

    @property
    def menu_entry(self) -> str:
        """Return menu entry representation of this deployment."""
        try:
            regions = self.regions or self.parallel_regions
        except UnresolvedVariable as err:
            LOGGER.debug(
                "attempted to use variable %s before it was resolved; "
                "using literal value in menu entry",
                err.variable.name,
            )
            regions = self._data.regions or self._data.parallel_regions
        return (
            f"{self.name} - {', '.join(module.name for module in self.modules)} "
            f"({', '.join(regions if isinstance(regions, list) else [regions])})"
        )

    @property
    def modules(self) -> List[RunwayModuleDefinition]:
        """List of Runway modules."""
        return [RunwayModuleDefinition(module) for module in self._data.modules]

    @modules.setter
    def modules(self, modules: List[RunwayModuleDefinition]) -> None:
        """Set the value of the property.

        Args:
            modules: A list of modules.

        Raises:
            TypeError: The provided value does not match the required types.

        """
        if not all(isinstance(i, RunwayModuleDefinition) for i in modules):  # type: ignore
            raise TypeError("modules must be type List[RunwayModuleDefinition]")
        self._data.modules = [
            RunwayModuleDefinitionModel.parse_obj(mod.data) for mod in modules
        ]

    def reverse(self):
        """Reverse the order of modules and regions."""
        self._data.modules.reverse()
        for mod in self._data.modules:
            mod.parallel.reverse()
        for prop in [self._data.parallel_regions, self._data.regions]:
            if isinstance(prop, list):
                prop.reverse()

    def set_modules(
        self, modules: List[Union[RunwayModuleDefinition, RunwayModuleDefinitionModel]]
    ) -> None:
        """Set the value of modules.

        Args:
            modules: A list of modules.

        Raises:
            TypeError: The provided value does not match the required types.

        """
        if not isinstance(modules, list):  # type: ignore
            raise TypeError(
                f"expected List[RunwayModuleDefinition]; got {type(modules)}"
            )
        sanitized: List[RunwayModuleDefinitionModel] = []
        for i, mod in enumerate(modules):
            if isinstance(mod, RunwayModuleDefinition):
                sanitized.append(RunwayModuleDefinitionModel.parse_obj(mod.data))
            elif isinstance(mod, RunwayModuleDefinitionModel):  # type: ignore
                sanitized.append(mod)
            else:
                raise TypeError(
                    f"{self.name}.modules[{i}] is type {type(mod)}; "
                    "expected type RunwayModuleDefinition or RunwayModuleDefinitionModel"
                )
        self._data.modules = sanitized

    def _register_variable(self, var_name: str, var_value: Any) -> None:
        """Register a variable.

        Args:
            var_name: Name of the config field that can contain a variable
                lookup.
            var_value: Literal value supplied in the config to be resolved
                as a variable if it contains a lookup.

        """
        self._vars[var_name] = Variable(
            name=f"{self.name}.{var_name}", value=var_value, variable_type="runway"
        )

    @overload
    @classmethod
    def parse_obj(cls, obj: List[Dict[str, Any]]) -> List[RunwayDeploymentDefinition]:
        ...

    @overload
    @classmethod
    def parse_obj(
        cls,
        obj: Union[
            List[ConfigProperty], Set[ConfigProperty], Tuple[ConfigProperty, ...]
        ],
    ) -> List[RunwayDeploymentDefinition]:
        ...

    @overload
    @classmethod
    def parse_obj(
        cls, obj: Union[Dict[str, Any], ConfigProperty]
    ) -> RunwayDeploymentDefinition:
        ...

    @classmethod
    def parse_obj(  # type: ignore
        cls, obj: Any
    ) -> Union[RunwayDeploymentDefinition, List[RunwayDeploymentDefinition]]:
        """Parse a python object into this class.

        Args:
            obj: The object to parse.

        """
        if isinstance(obj, (list, set, tuple)):
            return [
                cls(RunwayDeploymentDefinitionModel.parse_obj(o)) for o in obj  # type: ignore
            ]
        return cls(RunwayDeploymentDefinitionModel.parse_obj(obj))



"""Parse the given Cognito/Local authorization information.

Ensure the parsed Cognito and Local authorization information from the retrieved
query string parameter and cookie headers matches expectations. If it doesn't
then inform the user of a bad request, otherwise retrieve the Cognito tokens to
add to the cookie headers.
"""
# pylint: disable=consider-using-f-string
import base64
import hmac
import json
import logging
from datetime import datetime
from urllib.parse import parse_qs

from shared_jose import (  # noqa pylint: disable=import-error
    MissingRequiredGroupError,
    validate_and_check_id_token,
)

from shared import (  # noqa pylint: disable=import-error
    create_error_html,
    extract_and_parse_cookies,
    generate_cookie_headers,
    get_config,
    http_post_with_retry,
    sign,
    timestamp_in_seconds,
)

LOGGER = logging.getLogger(__name__)
CONFIG = get_config()
COGNITO_TOKEN_ENDPOINT = f"https://{CONFIG['cognito_auth_domain']}/oauth2/token"
NONCE_MAX_AGE = 60 * 60 * 24


class RequiresConfirmationError(Exception):
    """Raised when user needs to be sent back to Cognito."""


def validate_querystring_and_cookies(request, cookies):
    """Handle the authorization parsing.

    Args:
        request (Any): Cloudfront request.
        cookies (Dict[str, Any]): Cookies.

    """
    qsp = parse_qs(request.get("querystring"))
    # Authorization code given by Cognito
    code = qsp.get("code")
    # The requested URI and current nonce information
    state = json.loads(base64.urlsafe_b64decode(qsp.get("state")[0]).decode())

    if qsp.get("error"):
        raise Exception(f"[Cognito] {qsp['error']}: {qsp.get('error_description')}")

    # Missing required components
    if not code or not state:
        msg = (
            "Invalid query string. "
            'Your query string should include parameters "state" and "code" '
            "(this can be caused by authentication attempts originating from "
            "this site, which is not allowed)"
        )
        LOGGER.info(msg)
        raise Exception(msg)

    current_nonce = state.get("nonce")
    requested_uri = state.get("requestedUri", "")

    # Retrieve the original nonce value as well as our PKCE code
    original_nonce = cookies["nonce"]
    pkce = cookies["pkce"]
    nonce_hmac = cookies["nonce_hmac"]

    # If we're missing one of the nonces, or they don't match, cause an error
    if not current_nonce or not original_nonce or current_nonce != original_nonce:
        # No original nonce? CSRF violation
        if not original_nonce:
            msg = (
                "Your browser didn't send the nonce cookie along, "
                "but it is required for security (prevent CSRF)"
            )
            LOGGER.debug(msg)
            raise RequiresConfirmationError(msg)
        # Nonce's don't match
        msg = "Nonce mismatch (multiple parallel authentication attempts?)"
        LOGGER.debug(msg)
        raise RequiresConfirmationError(msg)
    if not pkce:
        raise Exception(
            "Your browser didn't send the pkce cookie along, "
            "but it is required for security (prevent CSRF)"
        )

    # Nonce should not be too old
    nonce_timestamp = int(current_nonce.split("T")[0])
    if (timestamp_in_seconds() - nonce_timestamp) > NONCE_MAX_AGE:
        raise RequiresConfirmationError(
            "Nonce is too old (nonce timestnonce_timestampamp is %s)"
            % datetime.fromtimestamp(nonce_timestamp).isoformat()
        )

    # Check nonce signature to ensure we generated it
    calculated_hmac = sign(current_nonce, CONFIG["nonce_signing_secret"])
    if not hmac.compare_digest(calculated_hmac, nonce_hmac):
        raise RequiresConfirmationError(
            "Nonce signature mismatch; expected %s but got %s"
            % (calculated_hmac, nonce_hmac)
        )

    return [code, pkce, requested_uri]


def handler(event, _context):
    """Handle the authorization parsing.

    Args:
        event (Any): The Lambda Event.
        _context (Any): Lambda context object.

    """
    request = event["Records"][0]["cf"]["request"]
    domain_name = request["headers"]["host"][0]["value"]
    redirected_from_uri = f"https://{domain_name}"
    id_token = None

    # Attempt to parse the request and retrieve authorization
    # tokens to integrate with our header cookies
    try:
        # Get all the cookies from the headers
        cookies = extract_and_parse_cookies(request.get("headers"), CONFIG["client_id"])
        id_token = cookies["id_token"]
        code, pkce, requested_uri = validate_querystring_and_cookies(request, cookies)
        redirected_from_uri += requested_uri

        # Request tokens from our Cognito Authorization Domain
        body = {
            "grant_type": "authorization_code",
            "client_id": CONFIG["client_id"],
            "redirect_uri": "https://%s%s"
            % (domain_name, CONFIG.get("redirect_path_sign_in")),
            "code": code[0],
            "code_verifier": pkce,
        }
        tokens = http_post_with_retry(
            COGNITO_TOKEN_ENDPOINT,
            body,
            {"Content-Type": "application/x-www-form-urlencoded"},
        )

        if not tokens:
            raise Exception("Was not able to obtain tokens from Cognito")

        # Validate the token information against the Cognito JWKS
        # (and ensure group membership, if applicable)
        validate_and_check_id_token(
            tokens["id_token"],
            CONFIG["token_jwks_uri"],
            CONFIG["token_issuer"],
            CONFIG["client_id"],
            CONFIG.get("required_group"),
        )

        # Redirect user to the originally requested uri with the
        # token header cookies
        response = {
            "status": "307",
            "statusDescription": "Temporary Redirect",
            "headers": {
                "location": [{"key": "location", "value": redirected_from_uri}],
                "set-cookie": generate_cookie_headers(
                    "new_tokens",
                    CONFIG.get("client_id"),
                    CONFIG.get("oauth_scopes"),
                    tokens,
                    domain_name,
                    CONFIG.get("cookie_settings"),
                ),
                **CONFIG.get("cloud_front_headers", {}),
            },
        }
        return response
    except Exception as err:  # pylint: disable=broad-except
        if id_token:
            # ID token found; checking if it is valid
            try:
                validate_and_check_id_token(
                    id_token,
                    CONFIG["token_jwks_uri"],
                    CONFIG["token_issuer"],
                    CONFIG["client_id"],
                    CONFIG.get("required_group"),
                )
                # Token is valid; return user to where they came from
                return {
                    "status": "307",
                    "statusDescription": "Temporary Redirect",
                    "headers": {
                        "location": [{"key": "location", "value": redirected_from_uri}],
                        **CONFIG.get("cloud_front_headers", {}),
                    },
                }
            except Exception as err2:  # pylint: disable=broad-except
                LOGGER.debug("Id token not valid")
                LOGGER.debug(err2)

        if isinstance(err, RequiresConfirmationError):
            html_params = [
                "Confirm sign-in",
                f"We need your confirmation to sign you ({err})",
                redirected_from_uri,
                "Confirm",
            ]
        elif isinstance(err, MissingRequiredGroupError):
            html_params = [
                "Not Authorized",
                "Your user is not authorized for this site. Please contact the admin.",
                redirected_from_uri,
                "Try Again",
            ]
        else:
            html_params = [
                "Sign-in issue",
                f"Sign-in unsuccessful because of a technical problem: {err}",
                redirected_from_uri,
                "Try Again",
            ]

        # Inform user of bad request and reason
        return {
            "body": create_error_html(*html_params),
            "status": "200",
            "headers": {
                **CONFIG.get("cloud_front_headers", {}),
                "content-type": [
                    {"key": "Content-Type", "value": "text/html; charset=UTF-8"}
                ],
            },
        }



"""``runway plan`` command."""
# docs: file://./../../../docs/source/commands.rst
import logging
from typing import Any, Tuple

import click
from pydantic import ValidationError

from ...core import Runway
from ...exceptions import ConfigNotFound, VariablesFileNotFound
from .. import options
from ..utils import select_deployments

LOGGER = logging.getLogger(__name__.replace("._", "."))


@click.command("plan", short_help="plan things")
@options.ci
@options.debug
@options.deploy_environment
@options.no_color
@options.tags
@options.verbose
@click.pass_context
def plan(ctx: click.Context, debug: bool, tags: Tuple[str, ...], **_: Any) -> None:
    """Determine what infrastructure changes will occur during the next deploy.

    \b
    Process
    -------
    1. Determines the deploy environment.
        - "-e, --deploy-environment" option
        - "DEPLOY_ENVIRONMENT" environment variable
        - git branch name
            - strips "ENV-" prefix, master is converted to common
            - ignored if "ignore_git_branch: true"
        - name of the current working directory
    2. Selects deployments & modules to deploy.
        - (default) prompts
        - (tags) module contains all tags
        - (non-interactive) all
    3. Attempt to determine change for deployments/modules in the order defined.

    """  # noqa: D301
    try:
        Runway(ctx.obj.runway_config, ctx.obj.get_runway_context()).plan(
            select_deployments(ctx, ctx.obj.runway_config.deployments, tags)
        )
    except ValidationError as err:
        LOGGER.error(err, exc_info=debug)
        ctx.exit(1)
    except (ConfigNotFound, VariablesFileNotFound) as err:
        LOGGER.error(err.message, exc_info=debug)
        ctx.exit(1)



"""Core Runway API."""
from __future__ import annotations

import logging as _logging
import sys as _sys
import traceback as _traceback
from typing import TYPE_CHECKING, Any, Dict, List, Optional, cast

import yaml as _yaml

from .. import __version__
from .._logging import PrefixAdaptor as _PrefixAdaptor
from .._logging import RunwayLogger as _RunwayLogger
from ..tests.registry import TEST_HANDLERS as _TEST_HANDLERS
from ..utils import DOC_SITE
from ..utils import YamlDumper as _YamlDumper
from . import components, providers, type_defs

if TYPE_CHECKING:
    from ..config import RunwayConfig
    from ..config.components.runway import RunwayDeploymentDefinition
    from ..context import RunwayContext

LOGGER = cast(_RunwayLogger, _logging.getLogger(__name__))

__all__ = ["Runway", "components", "providers", "type_defs"]


class Runway:
    """Runway's core functionality."""

    def __init__(self, config: RunwayConfig, context: RunwayContext) -> None:
        """Instantiate class.

        Args:
            config: Runway config.
            context: Runway context.

        """
        self.ctx = context
        self.deployments = config.deployments
        self.future = config.future
        self.required_version = config.runway_version
        self.tests = config.tests
        self.ignore_git_branch = config.ignore_git_branch
        self.variables = config.variables
        self.__assert_config_version()
        self.ctx.env.log_name()

    def deploy(
        self, deployments: Optional[List[RunwayDeploymentDefinition]] = None
    ) -> None:
        """Deploy action.

        Args:
            deployments: List of deployments to run. If not provided,
                all deployments in the config will be run.

        """
        self.__run_action(
            "deploy", deployments if deployments is not None else self.deployments
        )

    def destroy(
        self, deployments: Optional[List[RunwayDeploymentDefinition]] = None
    ) -> None:
        """Destroy action.

        Args:
            deployments: List of deployments to run. If not provided,
                all deployments in the config will be run in reverse.

        """
        self.__run_action(
            "destroy",
            deployments
            if deployments is not None
            else self.reverse_deployments(self.deployments),
        )
        if not deployments:
            # return config attribute to original state
            self.reverse_deployments(self.deployments)

    def get_env_vars(
        self, deployments: Optional[List[RunwayDeploymentDefinition]] = None
    ) -> Dict[str, Any]:
        """Get env_vars defined in the config.

        Args:
            deployments: List of deployments to get env_vars from.

        Returns:
            Resolved env_vars from the deployments.

        """
        deployments = deployments or self.deployments
        result: Dict[str, str] = {}
        for deployment in deployments:
            obj = components.Deployment(
                context=self.ctx, definition=deployment, variables=self.variables
            )
            result.update(obj.env_vars_config)
        return result

    def init(
        self, deployments: Optional[List[RunwayDeploymentDefinition]] = None
    ) -> None:
        """Init action.

        Args:
            deployments: List of deployments to run. If not provided,
                all deployments in the config will be run.

        """
        self.__run_action(
            "init", deployments if deployments is not None else self.deployments
        )

    def plan(
        self, deployments: Optional[List[RunwayDeploymentDefinition]] = None
    ) -> None:
        """Plan action.

        Args:
            deployments: List of deployments to run. If not provided,
                all deployments in the config will be run.

        """
        self.__run_action(
            "plan", deployments if deployments is not None else self.deployments
        )

    @staticmethod
    def reverse_deployments(
        deployments: List[RunwayDeploymentDefinition],
    ) -> List[RunwayDeploymentDefinition]:
        """Reverse deployments and the modules within them.

        Args:
            deployments: List of deployments to reverse.

        Returns:
            Deployments and modules in reverse order.

        """
        result: List[RunwayDeploymentDefinition] = []
        for deployment in deployments:
            deployment.reverse()
            result.insert(0, deployment)
        return result

    def test(self) -> None:
        """Run tests defined in the config."""
        if not self.tests:
            LOGGER.error("no tests defined in runway.yml")
            LOGGER.error(
                "to learn more about using Runway to run tests, visit "
                "%s/page/defining_tests.html",
                DOC_SITE,
            )
            LOGGER.error(
                "Example test:\n%s",
                _yaml.dump(
                    {
                        "tests": [
                            {
                                "name": "example-test",
                                "type": "script",
                                "required": True,
                                "args": {"commands": ['echo "Success!"']},
                            }
                        ]
                    },
                    Dumper=_YamlDumper,
                ),
            )
            _sys.exit(1)
        self.ctx.command = "test"

        failed_tests: List[str] = []

        LOGGER.info("found %i test(s)", len(self.tests))
        for tst in self.tests:
            tst.resolve(self.ctx, variables=self.variables)
            logger = _PrefixAdaptor(tst.name, LOGGER)
            logger.notice("running test (in progress)")
            try:
                handler = _TEST_HANDLERS[tst.type]
            except KeyError:
                logger.error('unable to find handler of type "%s"', tst.type)
                if tst.required:
                    _sys.exit(1)
                failed_tests.append(tst.name)
                continue
            try:
                handler.handle(tst.name, tst.args)
                logger.success("running test (pass)")
            except (Exception, SystemExit) as err:  # pylint: disable=broad-except
                # for lack of an easy, better way to do this atm, assume
                # SystemExits are due to a test failure and the failure reason
                # has already been properly logged by the handler or the
                # tool it is wrapping.
                if not isinstance(err, SystemExit):
                    _traceback.print_exc()
                elif err.code == 0:
                    continue  # tests with zero exit code don't indicate failure
                logger.error("running test (fail)")
                if tst.required:
                    logger.error("test required; the remaining tests have been skipped")
                    _sys.exit(1)
                failed_tests.append(tst.name)
        if failed_tests:
            LOGGER.error("the following tests failed: %s", ", ".join(failed_tests))
            _sys.exit(1)
        LOGGER.success("all tests passed")

    def __assert_config_version(self):
        """Assert the config supports this version of Runway."""
        if not self.required_version:
            LOGGER.debug("required Runway version not specified")
            return
        if __version__ in self.required_version:
            LOGGER.debug(
                'current Runway version "%s" matches "%s" required by this config file',
                __version__,
                self.required_version,
            )
            return
        if __version__.startswith("0.") and "dev" in __version__:
            LOGGER.warning(
                "Runway is being used from a shallow clone of the repo; "
                "config version will not be enforced as version cannot be determined"
            )
            return
        LOGGER.error(
            'current Runway version "%s" does not match "%s" required by this config file',
            __version__,
            self.required_version,
        )
        _sys.exit(1)

    def __run_action(
        self,
        action: type_defs.RunwayActionTypeDef,
        deployments: Optional[List[RunwayDeploymentDefinition]],
    ) -> None:
        """Run an action on a list of deployments.

        Args:
            action: Name of the action.
            deployments: List of deployments to run.

        """
        self.ctx.command = action
        components.Deployment.run_list(
            action=action,
            context=self.ctx,
            deployments=deployments or [],
            future=self.future,
            variables=self.variables,
        )



"""Install a version of kubectl."""
# docs: file://./../../../../docs/source/commands.rst
import logging
from typing import Any

import click

from ....env_mgr.kbenv import KBEnvManager
from ... import options

LOGGER = logging.getLogger(__name__.replace("._", "."))


@click.command("install", short_help="install kubectl")
@click.argument("version", metavar="[<version>]", required=False)
@options.debug
@options.no_color
@options.verbose
def install(version: str, **_: Any) -> None:
    """Install the specified <version> of kubectl (e.g. v1.14.0).

    If no version is specified, Runway will attempt to find and read a
    ".kubectl-version" file in the current directory.
    If this file doesn't exist, nothing will be installed.

    Compatible with https://github.com/alexppg/kbenv.

    """
    LOGGER.debug("kubectl path: %s", KBEnvManager().install(version_requested=version))



"""Install a version of Terraform."""
# docs: file://./../../../../docs/source/commands.rst
import logging
from typing import Any, Optional

import click

from ....env_mgr.tfenv import TFEnvManager
from ....utils import DOC_SITE
from ... import options

LOGGER = logging.getLogger(__name__.replace("._", "."))


@click.command("install", short_help="install terraform")
@click.argument("version", metavar="[<version>]", required=False, default=None)
@options.debug
@options.no_color
@options.verbose
@click.pass_context
def install(ctx: click.Context, version: Optional[str] = None, **_: Any) -> None:
    """Install the specified <version> of Terraform (e.g. 0.12.0).

    If no version is specified, Runway will attempt to find and read a
    ".terraform-version" file in the current directory.
    If this file doesn't exist, nothing will be installed.

    """
    try:
        LOGGER.debug(
            "terraform path: %s", TFEnvManager().install(version_requested=version)
        )
    except ValueError as err:
        LOGGER.debug("terraform install failed", exc_info=True)
        if "unable to find" not in str(err):
            LOGGER.error(
                "unexpected error encountered when trying to install Terraform",
                exc_info=True,
            )
            ctx.exit(1)
        else:
            LOGGER.error("unable to find a .terraform-version file")
            LOGGER.error(
                "learn how to use Runway to manage Terraform versions at "
                "%s/page/terraform/advanced_features.html#version-management",
                DOC_SITE,
            )
        ctx.exit(1)



"""Hello stack."""
from aws_cdk import aws_iam as iam
from aws_cdk import aws_sns as sns
from aws_cdk import aws_sns_subscriptions as subs
from aws_cdk import aws_sqs as sqs
from aws_cdk import core

from .hello_construct import HelloConstruct


class MyStack(core.Stack):
    """My stack."""

    def __init__(self, scope: core.Construct, id: str, **kwargs) -> None:
        """Instantiate class."""
        super().__init__(scope, id, **kwargs)

        queue = sqs.Queue(
            self, "MyFirstQueue", visibility_timeout=core.Duration.seconds(300)
        )

        topic = sns.Topic(self, "MyFirstTopic", display_name="My First Topic")

        topic.add_subscription(subs.SqsSubscription(queue))

        hello = HelloConstruct(self, "MyHelloConstruct", num_buckets=4)
        user = iam.User(self, "MyUser")
        hello.grant_read(user)



"""Base models & other objects."""
from __future__ import annotations

import pydantic

from ...utils import BaseModel


class ConfigProperty(BaseModel):
    """Base class for Runway configuration properties."""

    class Config(pydantic.BaseConfig):
        """Model configuration."""

        validate_all = True
        validate_assignment = True



"""Runway Kubernetes Module options."""
from __future__ import annotations

from pathlib import Path
from typing import Optional

from pydantic import Extra

from ...base import ConfigProperty


class RunwayK8sModuleOptionsDataModel(ConfigProperty):
    """Model for Runway Kubernetes Module options."""

    kubectl_version: Optional[str] = None
    overlay_path: Optional[Path] = None

    class Config(ConfigProperty.Config):
        """Model configuration."""

        extra = Extra.ignore
        title = "Runway Kubernetes Module options."



"""Cloudformation module."""
from __future__ import annotations

import logging
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Optional, Union, cast

from .._logging import PrefixAdaptor
from ..cfngin.cfngin import CFNgin
from .base import RunwayModule

if TYPE_CHECKING:
    from .._logging import RunwayLogger
    from ..context import RunwayContext
    from .base import ModuleOptions

LOGGER = cast("RunwayLogger", logging.getLogger(__name__))


class CloudFormation(RunwayModule):
    """CloudFormation (CFNgin) Runway Module."""

    def __init__(
        self,
        context: RunwayContext,
        *,
        explicitly_enabled: Optional[bool] = False,
        logger: RunwayLogger = LOGGER,
        module_root: Path,
        name: Optional[str] = None,
        options: Optional[Union[Dict[str, Any], ModuleOptions]] = None,
        parameters: Optional[Dict[str, Any]] = None,
        **_: Any,
    ) -> None:
        """Instantiate class.

        Args:
            context: Runway context object for the current session.
            explicitly_enabled: Whether or not the module is explicitly enabled.
                This is can be set in the event that the current environment being
                deployed to matches the defined environments of the module/deployment.
            logger: Used to write logs.
            module_root: Root path of the module.
            name: Name of the module.
            options: Options passed to the module class from the config as ``options``
                or ``module_options`` if coming from the deployment level.
            parameters: Values to pass to the underlying infrastructure as code
                tool that will alter the resulting infrastructure being deployed.
                Used to templatize IaC.

        """
        super().__init__(
            context,
            explicitly_enabled=explicitly_enabled,
            logger=logger,
            module_root=module_root,
            name=name,
            options=options,
            parameters=parameters,
        )
        # logger needs to be created here to use the correct logger
        self.logger = PrefixAdaptor(self.name, LOGGER)

    def deploy(self) -> None:
        """Run deploy."""
        cfngin = CFNgin(self.ctx, parameters=self.parameters, sys_path=self.path)
        cfngin.init(force=bool(self.parameters or self.explicitly_enabled))
        cfngin.deploy(force=bool(self.parameters or self.explicitly_enabled))

    def destroy(self) -> None:
        """Run destroy."""
        cfngin = CFNgin(self.ctx, parameters=self.parameters, sys_path=self.path)
        cfngin.destroy(force=bool(self.parameters or self.explicitly_enabled))

    def init(self) -> None:
        """Run init."""
        cfngin = CFNgin(self.ctx, parameters=self.parameters, sys_path=self.path)
        cfngin.init(force=bool(self.parameters or self.explicitly_enabled))

    def plan(self) -> None:
        """Run diff."""
        cfngin = CFNgin(self.ctx, parameters=self.parameters, sys_path=self.path)
        cfngin.plan(force=bool(self.parameters or self.explicitly_enabled))



"""Retrieve a value from SSM Parameter Store."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Union

from typing_extensions import Final, Literal

from ...lookups.handlers.base import LookupHandler

if TYPE_CHECKING:
    from ...context import CfnginContext, RunwayContext

LOGGER = logging.getLogger(__name__)


class SsmLookup(LookupHandler):
    """SSM Parameter Store Lookup."""

    TYPE_NAME: Final[Literal["ssm"]] = "ssm"
    """Name that the Lookup is registered as."""

    @classmethod
    def handle(  # pylint: disable=arguments-differ
        cls,
        value: str,
        context: Union[CfnginContext, RunwayContext],
        *__args: Any,
        **__kwargs: Any,
    ) -> Any:
        """Retrieve a value from SSM Parameter Store.

        Args:
            value: The value passed to the Lookup.
            context: The current context object.

        Raises:
            ParameterNotFound: Parameter not found in SSM and a default value
                was not provided.

        """
        query, args = cls.parse(value)

        session = context.get_session(region=args.get("region"))
        client = session.client("ssm")

        try:
            response = client.get_parameter(Name=query, WithDecryption=True)[
                "Parameter"
            ]
            return cls.format_results(
                response["Value"].split(",")
                if response["Type"] == "StringList"
                else response["Value"],
                **args,
            )
        except client.exceptions.ParameterNotFound:
            if args.get("default"):
                args.pop("load", None)  # don't load a default value
                return cls.format_results(args.pop("default"), **args)
            raise



"""'Git type Path Source."""
import logging
import shutil
import subprocess
import tempfile
from pathlib import Path
from typing import Any, Dict, Optional

from .source import Source

LOGGER = logging.getLogger(__name__)


class Git(Source):
    """Git Path Source.

    The Git path source can be tasked with cloning a remote repository
    and pointing to a specific module folder (or the root).

    """

    def __init__(
        self,
        *,
        arguments: Optional[Dict[str, str]] = None,
        location: str = "",
        uri: str = "",
        **kwargs: Any,
    ) -> None:
        """Git Path Source.

        Args:
            arguments: A reference can be passed along via the arguments so that a specific
                version of the repository is cloned. **commit**, **tag**, **branch**
                are all valid keys with respective output
            location: The relative location to the root of the repository where the
                module resides. Leaving this as an empty string, ``/``, or ``./``
                will have runway look in the root folder.
            uri: The uniform resource identifier that targets the remote git repository

        """
        self.args = arguments or {}
        self.uri = uri
        self.location = location

        super().__init__(**kwargs)

    def fetch(self) -> Path:
        """Retrieve the git repository from it's remote location."""
        from git.repo import Repo  # pylint: disable=import-outside-toplevel

        ref = self.__determine_git_ref()
        dir_name = "_".join([self.sanitize_git_path(self.uri), ref])
        cached_dir_path = self.cache_dir / dir_name

        if cached_dir_path.exists():
            return cached_dir_path

        with tempfile.TemporaryDirectory() as tmpdirname:
            tmp_repo_path = Path(tmpdirname) / dir_name
            with Repo.clone_from(self.uri, str(tmp_repo_path)) as repo:
                repo.head.set_reference(ref)
                repo.head.reset(index=True, working_tree=True)
            shutil.move(str(tmp_repo_path), self.cache_dir)

        return cached_dir_path

    def __git_ls_remote(self, ref: str) -> str:
        """List remote repositories based on uri and ref received.

        Keyword Args:
            ref (str): The git reference value

        """
        cmd = ["git", "ls-remote", self.uri, ref]
        LOGGER.debug("getting commit ID from repo: %s", " ".join(cmd))
        ls_remote_output = subprocess.check_output(cmd)
        if b"\t" in ls_remote_output:
            commit_id = ls_remote_output.split(b"\t", maxsplit=1)[0].decode()
            LOGGER.debug("matching commit id found: %s", commit_id)
            return commit_id
        raise ValueError(f'Ref "{ref}" not found for repo {self.uri}.')

    def __determine_git_ls_remote_ref(self) -> str:
        """Determine remote ref, defaulting to HEAD unless a branch is found."""
        ref = "HEAD"
        if self.args.get("branch"):
            ref = f"refs/heads/{self.args['branch']}"
        return ref

    def __determine_git_ref(self) -> str:
        """Determine the git reference code."""
        ref_config_keys = sum(
            bool(self.args.get(i)) for i in ["commit", "tag", "branch"]
        )
        if ref_config_keys > 1:
            raise ValueError(
                "Fetching remote git sources failed: conflicting revisions "
                "(e.g. 'commit', 'tag', 'branch') specified for a package source"
            )

        if self.args.get("commit"):
            return self.args["commit"]
        if self.args.get("tag"):
            return self.args["tag"]
        return self.__git_ls_remote(self.__determine_git_ls_remote_ref())

    @classmethod
    def sanitize_git_path(cls, path: str) -> str:
        """Sanitize the git path for folder/file assignment.

        Keyword Args:
            path: The path string to be sanitized

        """
        dir_name = path
        split = path.split("//")
        domain = split[len(split) - 1]

        if domain.endswith(".git"):
            dir_name = domain[:-4]

        return cls.sanitize_directory_path(dir_name)



"""S3 handler.

.. Derived from software distributed by Amazon.com, Inc - http://aws.amazon.com/apache2.0/
   https://github.com/aws/aws-cli/blob/83b43782dd/awscli/customizations/s3/s3handler.py

"""
from __future__ import annotations

import logging
import os
import sys
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    Type,
    cast,
)

from s3transfer.manager import TransferManager

from .results import (
    CommandResultRecorder,
    CopyResultSubscriber,
    DeleteResultSubscriber,
    DownloadResultSubscriber,
    DownloadStreamResultSubscriber,
    DryRunResult,
    FailureResult,
    NoProgressResultPrinter,
    OnlyShowErrorsResultPrinter,
    QueuedResult,
    ResultPrinter,
    ResultProcessor,
    ResultRecorder,
    SuccessResult,
    Union,
    UploadResultSubscriber,
    UploadStreamResultSubscriber,
)
from .transfer_config import create_transfer_config_from_runtime_config
from .utils import (
    MAX_UPLOAD_SIZE,
    DeleteCopySourceObjectSubscriber,
    DeleteSourceFileSubscriber,
    DeleteSourceObjectSubscriber,
    DirectoryCreatorSubscriber,
    NonSeekableStream,
    ProvideCopyContentTypeSubscriber,
    ProvideLastModifiedTimeSubscriber,
    ProvideSizeSubscriber,
    ProvideUploadContentTypeSubscriber,
    RequestParamsMapper,
    StdoutBytesWriter,
    create_warning,
    find_bucket_key,
    human_readable_size,
    relative_path,
)

if TYPE_CHECKING:
    from queue import Queue

    from mypy_boto3_s3.client import S3Client
    from s3transfer.futures import TransferFuture
    from s3transfer.subscribers import BaseSubscriber

    from ......type_defs import AnyPath
    from .file_info import FileInfo
    from .parameters import ParametersDataModel
    from .results import CommandResult
    from .transfer_config import TransferConfigDict

LOGGER = logging.getLogger(__name__.replace("._", "."))


class StdinMissingError(Exception):
    """Stdin missing."""

    def __init__(self) -> None:
        """Instantiate class."""
        self.message = "stdin is required for this operation, but is not available"
        super().__init__(self.message)


class S3TransferHandlerFactory:
    """Factory for S3TransferHandlers."""

    MAX_IN_MEMORY_CHUNKS: ClassVar[int] = 6

    def __init__(
        self, config_params: ParametersDataModel, runtime_config: TransferConfigDict
    ) -> None:
        """Instantiate class.

        Args:
            config_params: The parameters provide to the CLI command.
            runtime_config: The runtime config for the CLI command
                being run.

        """
        self._config_params = config_params
        self._runtime_config = runtime_config

    def __call__(
        self, client: S3Client, result_queue: "Queue[Any]"
    ) -> S3TransferHandler:
        """Create a S3TransferHandler instance.

        Args:
            client: The client to power the S3TransferHandler.
            result_queue: The result queue to be used to process results
                for the S3TransferHandler.

        """
        transfer_config = create_transfer_config_from_runtime_config(
            self._runtime_config
        )
        transfer_config.max_in_memory_upload_chunks = self.MAX_IN_MEMORY_CHUNKS
        transfer_config.max_in_memory_download_chunks = self.MAX_IN_MEMORY_CHUNKS

        transfer_manager = TransferManager(client, transfer_config)

        LOGGER.debug(
            "Using a multipart threshold of %s and a part size of %s",
            transfer_config.multipart_threshold,
            transfer_config.multipart_chunksize,
        )
        result_recorder = ResultRecorder()
        result_processor_handlers: List[Any] = [result_recorder]
        self._add_result_printer(result_recorder, result_processor_handlers)
        result_processor = ResultProcessor(
            result_queue=result_queue, result_handlers=result_processor_handlers
        )
        command_result_recorder = CommandResultRecorder(
            result_queue=result_queue,
            result_recorder=result_recorder,
            result_processor=result_processor,
        )

        return S3TransferHandler(
            transfer_manager=transfer_manager,
            config_params=self._config_params,
            result_command_recorder=command_result_recorder,
        )

    def _add_result_printer(
        self,
        result_recorder: ResultRecorder,
        result_processor_handlers: List[
            Union[
                NoProgressResultPrinter,
                OnlyShowErrorsResultPrinter,
                ResultPrinter,
                ResultRecorder,
            ]
        ],
    ) -> None:
        if self._config_params.quiet:
            return
        if self._config_params.only_show_errors:
            result_printer = OnlyShowErrorsResultPrinter(result_recorder)
        elif self._config_params.is_stream:
            result_printer = OnlyShowErrorsResultPrinter(result_recorder)
        elif self._config_params.no_progress:
            result_printer = NoProgressResultPrinter(result_recorder)
        else:
            result_printer = ResultPrinter(result_recorder)
        result_processor_handlers.append(result_printer)


class S3TransferHandler:
    """Backend for performing S3 transfers."""

    def __init__(
        self,
        transfer_manager: TransferManager,
        config_params: ParametersDataModel,
        result_command_recorder: CommandResultRecorder,
    ) -> None:
        """Instantiate class.

        Args:
            transfer_manager: Transfer manager to use for transfers
            config_params: The parameters passed to the CLI command in the
                form of a dictionary
            result_command_recorder: The result command recorder to be
                used to get the final result of the transfer

        """
        self._transfer_manager = transfer_manager
        self._result_command_recorder = result_command_recorder

        submitter_args = (
            self._transfer_manager,
            self._result_command_recorder.result_queue,
            config_params,
        )
        self._submitters = [
            UploadStreamRequestSubmitter(*submitter_args),
            DownloadStreamRequestSubmitter(*submitter_args),
            UploadRequestSubmitter(*submitter_args),
            DownloadRequestSubmitter(*submitter_args),
            CopyRequestSubmitter(*submitter_args),
            DeleteRequestSubmitter(*submitter_args),
            LocalDeleteRequestSubmitter(*submitter_args),
        ]

    def call(self, fileinfos: Iterator[FileInfo]) -> CommandResult:
        """Process iterable of FileInfos for transfer.

        Args:
            fileinfos: Set of FileInfos to submit to underlying transfer
                request submitters to make transfer API calls to S3

        Returns:
            The result of the command that specifies the number of
            failures and warnings encountered.

        """
        with self._result_command_recorder:
            with self._transfer_manager:
                total_submissions = 0
                for fileinfo in fileinfos:
                    for submitter in self._submitters:
                        if submitter.can_submit(fileinfo):
                            if submitter.submit(fileinfo):
                                total_submissions += 1
                            break
                self._result_command_recorder.notify_total_submissions(
                    total_submissions
                )
        return self._result_command_recorder.get_command_result()


class BaseTransferRequestSubmitter:
    """Submits transfer requests to the TransferManager.

    Given a FileInfo object and provided CLI parameters, it will add the
    necessary extra arguments and subscribers in making a call to the
    TransferManager.

    """

    REQUEST_MAPPER_METHOD: ClassVar[
        Optional[Callable[[Dict[Any, Any], Dict[Any, Any]], Any]]
    ] = None
    RESULT_SUBSCRIBER_CLASS: ClassVar[Optional[Type[BaseSubscriber]]] = None

    def __init__(
        self,
        transfer_manager: TransferManager,
        result_queue: "Queue[Any]",
        config_params: ParametersDataModel,
    ):
        """Instantiate class.

        Args:
            transfer_manager: The underlying transfer manager.
            result_queue: The result queue to use.
            config_params: The associated CLI parameters passed in to the
                command as a dictionary.

        """
        self._transfer_manager = transfer_manager
        self._result_queue = result_queue
        self._config_params = config_params

    def submit(self, fileinfo: FileInfo) -> Optional[TransferFuture]:
        """Submit a transfer request based on the FileInfo provided.

        There is no guarantee that the transfer request will be made on
        behalf of the fileinfo as a fileinfo may be skipped based on
        circumstances in which the transfer is not possible.

        Args:
            fileinfo: The FileInfo to be used to submit a transfer
                request to the underlying transfer manager.

        Returns:
            A TransferFuture representing the transfer if it the
            transfer was submitted. If it was not submitted nothing
            is returned.

        """
        should_skip = self._warn_and_signal_if_skip(fileinfo)
        if not should_skip:
            return self._do_submit(fileinfo)
        return None

    def can_submit(self, fileinfo: FileInfo) -> bool:
        """Check whether it can submit a particular FileInfo.

        Args:
            fileinfo: The FileInfo to check if the transfer request
                submitter can handle.

        Returns:
            True if it can use the provided FileInfo to make a transfer
            request to the underlying transfer manager. False, otherwise.

        """
        raise NotImplementedError("can_submit()")

    def _do_submit(self, fileinfo: FileInfo) -> Optional[TransferFuture]:
        """Do submit."""
        extra_args: Dict[Any, Any] = {}
        if self.REQUEST_MAPPER_METHOD:
            # pylint: disable=not-callable
            # TODO revisit in future releases of pyright - not seeing second arg
            self.REQUEST_MAPPER_METHOD(extra_args, self._config_params.dict())  # type: ignore
        subscribers: List[BaseSubscriber] = []
        self._add_additional_subscribers(subscribers, fileinfo)
        # The result subscriber class should always be the last registered
        # subscriber to ensure it is not missing any information that
        # may have been added in a different subscriber such as size.
        if self.RESULT_SUBSCRIBER_CLASS:
            result_kwargs: Dict[str, Any] = {"result_queue": self._result_queue}
            if self._config_params.is_move:
                result_kwargs["transfer_type"] = "move"
            # pylint: disable=not-callable
            subscribers.append(self.RESULT_SUBSCRIBER_CLASS(**result_kwargs))

        if not self._config_params.dryrun:
            return self._submit_transfer_request(fileinfo, extra_args, subscribers)
        return self._submit_dryrun(fileinfo)

    def _submit_dryrun(self, fileinfo: FileInfo) -> None:
        """Submit dryrun."""
        transfer_type = fileinfo.operation_name
        if self._config_params.is_move:
            transfer_type = "move"
        src, dest = self._format_src_dest(fileinfo)
        self._result_queue.put(
            DryRunResult(transfer_type=transfer_type, src=src, dest=dest)
        )

    def _add_additional_subscribers(
        self, subscribers: List[BaseSubscriber], fileinfo: FileInfo
    ) -> None:
        """Add additional subscribers."""

    def _submit_transfer_request(
        self,
        fileinfo: FileInfo,
        extra_args: Dict[str, Any],
        subscribers: List[BaseSubscriber],
    ) -> TransferFuture:
        """Submit transfer request."""
        raise NotImplementedError("_submit_transfer_request()")

    def _warn_and_signal_if_skip(self, fileinfo: FileInfo) -> bool:
        """Warn and signal if skip."""
        for warning_handler in self._get_warning_handlers():
            if warning_handler(fileinfo):
                # On the first warning handler that returns a signal to skip
                # immediately propagate this signal and no longer check
                # the other warning handlers as no matter what the file will
                # be skipped.
                return True
        return False

    def _get_warning_handlers(self) -> List[Callable[[FileInfo], Any]]:
        """Return a list of warning handlers, which are callables.

        Handlers take in a single parameter representing a FileInfo.
        It will then add a warning to result_queue if needed and return True if
        that FileInfo should be skipped.

        """
        return []

    def _should_inject_content_type(self) -> bool:
        """If should inject content type."""
        return bool(
            self._config_params.guess_mime_type and not self._config_params.content_type
        )

    def _warn_glacier(self, fileinfo: FileInfo) -> bool:
        """Warn glacier."""
        if not self._config_params.force_glacier_transfer:
            if not fileinfo.is_glacier_compatible:
                LOGGER.debug(
                    "Encountered glacier object s3://%s. Not performing "
                    "%s on object.",
                    fileinfo.src,
                    fileinfo.operation_name,
                )
                if not self._config_params.ignore_glacier_warnings:
                    warning = create_warning(
                        f"s3://{fileinfo.src}",
                        "Object is of storage class GLACIER. Unable to "
                        f"perform {fileinfo.operation_name} operations on GLACIER objects. "
                        "You must restore the object to be able to perform the "
                        f"operation. See aws s3 {fileinfo.operation_name} help "
                        "for additional parameter options to ignore or force these "
                        "transfers.",
                    )
                    self._result_queue.put(warning)
                return True
        return False

    def _warn_parent_reference(self, fileinfo: FileInfo) -> bool:
        """Warn parent reference."""
        # normpath() will use the OS path separator so we
        # need to take that into account when checking for a parent prefix.
        parent_prefix = ".." + os.path.sep
        escapes_cwd = (
            os.path.normpath(fileinfo.compare_key).startswith(parent_prefix)
            if fileinfo.compare_key
            else False
        )
        if escapes_cwd:
            warning = create_warning(
                fileinfo.compare_key, "File references a parent directory."
            )
            self._result_queue.put(warning)
            return True
        return False

    def _format_src_dest(
        self, fileinfo: FileInfo
    ) -> Tuple[Optional[str], Optional[str]]:
        """Return formatted versions of a fileinfos source and destination."""
        raise NotImplementedError("_format_src_dest()")

    def _format_local_path(self, path: Optional[AnyPath]) -> Optional[str]:
        """Format local path."""
        return relative_path(path)

    def _format_s3_path(self, path: Optional[AnyPath]) -> Optional[str]:
        """Format s3 path."""
        if not path:
            return None
        path = str(path)
        if path.startswith("s3://"):
            return path
        return "s3://" + path


class UploadRequestSubmitter(BaseTransferRequestSubmitter):
    """Upload request submitter."""

    REQUEST_MAPPER_METHOD: ClassVar[
        Callable[[Dict[Any, Any], Dict[Any, Any]], Any]
    ] = RequestParamsMapper.map_put_object_params
    RESULT_SUBSCRIBER_CLASS: ClassVar[
        Type[UploadResultSubscriber]
    ] = UploadResultSubscriber

    def can_submit(self, fileinfo: FileInfo) -> bool:
        """Check whether it can submit a particular FileInfo.

        Args:
            fileinfo: The FileInfo to check if the transfer request
                submitter can handle.

        Returns:
            True if it can use the provided FileInfo to make a transfer
            request to the underlying transfer manager. False, otherwise.

        """
        return fileinfo.operation_name == "upload"

    def _add_additional_subscribers(
        self, subscribers: List[BaseSubscriber], fileinfo: FileInfo
    ) -> None:
        """Add additional subscribers."""
        subscribers.append(ProvideSizeSubscriber(fileinfo.size))
        if self._should_inject_content_type():
            subscribers.append(ProvideUploadContentTypeSubscriber())
        if self._config_params.is_move:
            subscribers.append(DeleteSourceFileSubscriber())

    def _submit_transfer_request(
        self,
        fileinfo: FileInfo,
        extra_args: Dict[str, Any],
        subscribers: List[BaseSubscriber],
    ) -> TransferFuture:
        """Submit transfer request."""
        bucket, key = find_bucket_key(str(fileinfo.dest))
        filein = self._get_filein(fileinfo)
        return self._transfer_manager.upload(
            fileobj=filein,
            bucket=bucket,
            key=key,
            extra_args=extra_args,
            subscribers=subscribers,
        )

    @staticmethod
    def _get_filein(fileinfo: FileInfo) -> str:
        """Get file in."""
        return str(fileinfo.src)

    def _get_warning_handlers(self) -> List[Callable[[FileInfo], Any]]:
        """Get warning handlers."""
        return [self._warn_if_too_large]

    def _warn_if_too_large(self, fileinfo: FileInfo) -> None:
        """Warn if too large."""
        if fileinfo.size and fileinfo.size > MAX_UPLOAD_SIZE:
            file_path = relative_path(fileinfo.src)
            warning_message = (
                f"File {file_path} exceeds s3 upload limit of "
                f"{human_readable_size(MAX_UPLOAD_SIZE)}."
            )
            warning = create_warning(file_path, warning_message, skip_file=False)
            self._result_queue.put(warning)

    def _format_src_dest(
        self, fileinfo: FileInfo
    ) -> Tuple[Optional[str], Optional[str]]:
        """Return formatted versions of a fileinfos source and destination."""
        src = self._format_local_path(fileinfo.src)
        dest = self._format_s3_path(fileinfo.dest)
        return src, dest


class DownloadRequestSubmitter(BaseTransferRequestSubmitter):
    """Download request submitter."""

    REQUEST_MAPPER_METHOD: ClassVar[
        Callable[[Dict[Any, Any], Dict[Any, Any]], Any]
    ] = RequestParamsMapper.map_get_object_params
    RESULT_SUBSCRIBER_CLASS: ClassVar[
        Type[DownloadResultSubscriber]
    ] = DownloadResultSubscriber

    def can_submit(self, fileinfo: FileInfo) -> bool:
        """Check whether it can submit a particular FileInfo.

        Args:
            fileinfo: The FileInfo to check if the transfer request
                submitter can handle.

        Returns:
            True if it can use the provided FileInfo to make a transfer
            request to the underlying transfer manager. False, otherwise.

        """
        return fileinfo.operation_name == "download"

    def _add_additional_subscribers(
        self, subscribers: List[BaseSubscriber], fileinfo: FileInfo
    ) -> None:
        """Add additional subscribers."""
        subscribers.append(ProvideSizeSubscriber(fileinfo.size))
        subscribers.append(DirectoryCreatorSubscriber())
        subscribers.append(
            ProvideLastModifiedTimeSubscriber(fileinfo.last_update, self._result_queue)
        )
        if self._config_params.is_move:
            subscribers.append(
                DeleteSourceObjectSubscriber(fileinfo.source_client)  # type: ignore
            )

    def _submit_transfer_request(
        self,
        fileinfo: FileInfo,
        extra_args: Dict[str, Any],
        subscribers: List[BaseSubscriber],
    ) -> TransferFuture:
        """Submit transfer request."""
        bucket, key = find_bucket_key(str(fileinfo.src))
        return self._transfer_manager.download(
            fileobj=self._get_fileout(fileinfo),
            bucket=bucket,
            key=key,
            extra_args=extra_args,
            subscribers=subscribers,
        )

    @staticmethod
    def _get_fileout(fileinfo: FileInfo) -> str:
        """Get file out."""
        return str(fileinfo.dest)

    def _get_warning_handlers(self) -> List[Callable[[FileInfo], Any]]:
        """Get warning handlers."""
        return [self._warn_glacier, self._warn_parent_reference]

    def _format_src_dest(
        self, fileinfo: FileInfo
    ) -> Tuple[Optional[str], Optional[str]]:
        """Return formatted versions of a fileinfos source and destination."""
        src = self._format_s3_path(fileinfo.src)
        dest = self._format_local_path(fileinfo.dest)
        return src, dest


class CopyRequestSubmitter(BaseTransferRequestSubmitter):
    """Copy request submitter."""

    REQUEST_MAPPER_METHOD: ClassVar[
        Callable[[Dict[Any, Any], Dict[Any, Any]], Any]
    ] = RequestParamsMapper.map_copy_object_params
    RESULT_SUBSCRIBER_CLASS: ClassVar[Type[CopyResultSubscriber]] = CopyResultSubscriber

    def can_submit(self, fileinfo: FileInfo) -> bool:
        """Check whether it can submit a particular FileInfo.

        Args:
            fileinfo: The FileInfo to check if the transfer request
                submitter can handle.

        Returns:
            True if it can use the provided FileInfo to make a transfer
            request to the underlying transfer manager. False, otherwise.

        """
        return fileinfo.operation_name == "copy"

    def _add_additional_subscribers(
        self, subscribers: List[BaseSubscriber], fileinfo: FileInfo
    ) -> None:
        """Add additional subscribers."""
        subscribers.append(ProvideSizeSubscriber(fileinfo.size))
        if self._should_inject_content_type():
            subscribers.append(ProvideCopyContentTypeSubscriber())
        if self._config_params.is_move:
            subscribers.append(
                DeleteCopySourceObjectSubscriber(fileinfo.source_client)  # type: ignore
            )

    def _submit_transfer_request(
        self,
        fileinfo: FileInfo,
        extra_args: Dict[str, Any],
        subscribers: List[BaseSubscriber],
    ) -> TransferFuture:
        """Submit transfer request."""
        bucket, key = find_bucket_key(str(fileinfo.dest))
        source_bucket, source_key = find_bucket_key(str(fileinfo.src))
        copy_source = {"Bucket": source_bucket, "Key": source_key}
        return self._transfer_manager.copy(
            bucket=bucket,
            key=key,
            copy_source=copy_source,
            extra_args=extra_args,
            subscribers=subscribers,
            source_client=cast("S3Client", fileinfo.source_client),
        )

    def _get_warning_handlers(self) -> List[Callable[[FileInfo], Any]]:
        """Get warning handlers."""
        return [self._warn_glacier]

    def _format_src_dest(
        self, fileinfo: FileInfo
    ) -> Tuple[Optional[str], Optional[str]]:
        """Return formatted versions of a fileinfos source and destination."""
        src = self._format_s3_path(fileinfo.src)
        dest = self._format_s3_path(fileinfo.dest)
        return src, dest


class UploadStreamRequestSubmitter(UploadRequestSubmitter):
    """Upload stream request submitter."""

    RESULT_SUBSCRIBER_CLASS: ClassVar[
        Type[UploadStreamResultSubscriber]
    ] = UploadStreamResultSubscriber

    def can_submit(self, fileinfo: FileInfo) -> bool:
        """Check whether it can submit a particular FileInfo.

        Args:
            fileinfo: The FileInfo to check if the transfer request
                submitter can handle.

        Returns:
            True if it can use the provided FileInfo to make a transfer
            request to the underlying transfer manager. False, otherwise.

        """
        return bool(
            fileinfo.operation_name == "upload" and self._config_params.is_stream
        )

    def _add_additional_subscribers(
        self, subscribers: List[BaseSubscriber], fileinfo: FileInfo
    ) -> None:
        """Add additional subscribers."""
        expected_size = self._config_params.expected_size
        if expected_size is not None:
            subscribers.append(ProvideSizeSubscriber(int(expected_size)))

    @staticmethod
    def _get_filein(fileinfo: FileInfo) -> NonSeekableStream:  # type: ignore
        """Get file in."""
        if sys.stdin is None:
            raise StdinMissingError()
        return NonSeekableStream(sys.stdin.buffer)

    def _format_local_path(self, path: Optional[AnyPath]) -> str:
        """Format local path."""
        return "-"


class DownloadStreamRequestSubmitter(DownloadRequestSubmitter):
    """Download stream result subscriber."""

    RESULT_SUBSCRIBER_CLASS: ClassVar[
        Type[DownloadStreamResultSubscriber]
    ] = DownloadStreamResultSubscriber

    def can_submit(self, fileinfo: FileInfo) -> bool:
        """Check whether it can submit a particular FileInfo.

        Args:
            fileinfo: The FileInfo to check if the transfer request
                submitter can handle.

        Returns:
            True if it can use the provided FileInfo to make a transfer
            request to the underlying transfer manager. False, otherwise.

        """
        return bool(
            fileinfo.operation_name == "download" and self._config_params.is_stream
        )

    def _add_additional_subscribers(
        self, subscribers: List[BaseSubscriber], fileinfo: FileInfo
    ) -> None:
        """Add additional subscribers."""

    @staticmethod
    def _get_fileout(fileinfo: FileInfo) -> StdoutBytesWriter:  # type: ignore
        """Get file out."""
        return StdoutBytesWriter()

    def _format_local_path(self, path: Optional[AnyPath]) -> str:
        """Format local path."""
        return "-"


class DeleteRequestSubmitter(BaseTransferRequestSubmitter):
    """Delete request submitter."""

    REQUEST_MAPPER_METHOD: ClassVar[
        Callable[[Dict[Any, Any], Dict[Any, Any]], Any]
    ] = RequestParamsMapper.map_delete_object_params
    RESULT_SUBSCRIBER_CLASS: ClassVar[
        Type[DeleteResultSubscriber]
    ] = DeleteResultSubscriber

    def can_submit(self, fileinfo: FileInfo) -> bool:
        """Check whether it can submit a particular FileInfo.

        Args:
            fileinfo: The FileInfo to check if the transfer request
                submitter can handle.

        Returns:
            True if it can use the provided FileInfo to make a transfer
            request to the underlying transfer manager. False, otherwise.

        """
        return fileinfo.operation_name == "delete" and fileinfo.src_type == "s3"

    def _submit_transfer_request(
        self,
        fileinfo: FileInfo,
        extra_args: Dict[str, Any],
        subscribers: List[BaseSubscriber],
    ) -> TransferFuture:
        """Submit transfer request."""
        bucket, key = find_bucket_key(str(fileinfo.src))
        return self._transfer_manager.delete(
            bucket=bucket, key=key, extra_args=extra_args, subscribers=subscribers
        )

    def _format_src_dest(
        self, fileinfo: FileInfo
    ) -> Tuple[Optional[str], Optional[str]]:
        """Return formatted versions of a fileinfos source and destination."""
        return self._format_s3_path(fileinfo.src), None


class LocalDeleteRequestSubmitter(BaseTransferRequestSubmitter):
    """Local delete request submitter."""

    REQUEST_MAPPER_METHOD: ClassVar[
        Optional[Callable[[Dict[Any, Any], Dict[Any, Any]], Any]]
    ] = None
    RESULT_SUBSCRIBER_CLASS: ClassVar[Optional[Type[BaseSubscriber]]] = None

    def can_submit(self, fileinfo: FileInfo) -> bool:
        """Check whether it can submit a particular FileInfo.

        Args:
            fileinfo: The FileInfo to check if the transfer request
                submitter can handle.

        Returns:
            True if it can use the provided FileInfo to make a transfer
            request to the underlying transfer manager. False, otherwise.

        """
        return fileinfo.operation_name == "delete" and fileinfo.src_type == "local"

    def _submit_transfer_request(  # type: ignore
        self,
        fileinfo: FileInfo,
        extra_args: Dict[str, Any],
        subscribers: List[BaseSubscriber],
    ) -> bool:
        """Submit transfer request.

        This is quirky but essentially instead of relying on a built-in
        method of s3 transfer, the logic lives directly in the submitter.
        The reason a explicit delete local file does not
        live in s3transfer is because it is outside the scope of s3transfer;
        it should only have interfaces for interacting with S3. Therefore,
        the burden of this functionality should live in the CLI.

        The main downsides in doing this is that delete and the result
        creation happens in the main thread as opposed to a separate thread
        in s3transfer. However, this is not too big of a downside because
        deleting a local file only happens for sync --delete downloads and
        is very fast compared to all of the other types of transfers.

        """
        src, dest = self._format_src_dest(fileinfo)
        result_kwargs = {"transfer_type": "delete", "src": src, "dest": dest}
        try:
            self._result_queue.put(QueuedResult(total_transfer_size=0, **result_kwargs))
            os.remove(fileinfo.src)
            self._result_queue.put(SuccessResult(**result_kwargs))
        except Exception as exc:  # pylint: disable=broad-except
            self._result_queue.put(FailureResult(exception=exc, **result_kwargs))
        return True

    def _format_src_dest(
        self, fileinfo: FileInfo
    ) -> Tuple[Optional[str], Optional[str]]:
        return self._format_local_path(fileinfo.src), None



"""Set package version."""
import logging
import sys

from ._logging import LogLevels, RunwayLogger  # noqa: F401

logging.setLoggerClass(RunwayLogger)

if sys.version_info < (3, 8):
    # importlib.metadata is standard lib for python>=3.8, use backport
    from importlib_metadata import PackageNotFoundError, version  # type: ignore
else:
    from importlib.metadata import PackageNotFoundError, version  # type: ignore

try:
    __version__ = version(__name__)
except PackageNotFoundError:
    # package is not installed
    __version__ = "0.0.0"



"""CFNgin diff action."""
from __future__ import annotations

import logging
import sys
from operator import attrgetter
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Tuple,
    TypeVar,
    Union,
    cast,
)

from botocore.exceptions import ClientError

from ...core.providers.aws.s3 import Bucket
from .. import exceptions
from ..status import (
    COMPLETE,
    INTERRUPTED,
    DoesNotExistInCloudFormation,
    NotSubmittedStatus,
    NotUpdatedStatus,
    SkippedStatus,
)
from . import deploy
from .base import build_walker

if TYPE_CHECKING:
    from ..._logging import RunwayLogger
    from ..stack import Stack
    from ..status import Status

_NV = TypeVar("_NV")
_OV = TypeVar("_OV")

LOGGER = cast("RunwayLogger", logging.getLogger(__name__))


class DictValue(Generic[_OV, _NV]):
    """Used to create a diff of two dictionaries."""

    ADDED = "ADDED"
    REMOVED = "REMOVED"
    MODIFIED = "MODIFIED"
    UNMODIFIED = "UNMODIFIED"

    formatter = "%s%s = %s"

    def __init__(self, key: str, old_value: _OV, new_value: _NV) -> None:
        """Instantiate class."""
        self.key = key
        self.old_value = old_value
        self.new_value = new_value

    def __eq__(self, other: object) -> bool:
        """Compare if self is equal to another object."""
        return self.__dict__ == other.__dict__

    def changes(self) -> List[str]:
        """Return changes to represent the diff between old and new value.

        Returns:
            Representation of the change (if any) between old and new value.

        """
        output: List[str] = []
        if self.status() is self.UNMODIFIED:
            output = [self.formatter % (" ", self.key, self.old_value)]
        elif self.status() is self.ADDED:
            output.append(self.formatter % ("+", self.key, self.new_value))
        elif self.status() is self.REMOVED:
            output.append(self.formatter % ("-", self.key, self.old_value))
        elif self.status() is self.MODIFIED:
            output.append(self.formatter % ("-", self.key, self.old_value))
            output.append(self.formatter % ("+", self.key, self.new_value))
        return output

    def status(self) -> str:
        """Status of changes between the old value and new value."""
        if self.old_value == self.new_value:
            return self.UNMODIFIED
        if self.old_value is None:
            return self.ADDED
        if self.new_value is None:
            return self.REMOVED
        return self.MODIFIED


def diff_dictionaries(
    old_dict: Dict[str, _OV], new_dict: Dict[str, _NV]
) -> Tuple[int, List[DictValue[_OV, _NV]]]:
    """Calculate the diff two single dimension dictionaries.

    Args:
        old_dict: Old dictionary.
        new_dict: New dictionary.

    Returns:
        Number of changed records and the :class:`DictValue` object containing
        the changes.

    """
    old_set = set(old_dict)
    new_set = set(new_dict)

    added_set = new_set - old_set
    removed_set = old_set - new_set
    common_set = old_set & new_set

    changes = 0
    output: List[DictValue[Any, Any]] = []
    for key in added_set:
        changes += 1
        output.append(DictValue(key, None, new_dict[key]))

    for key in removed_set:
        changes += 1
        output.append(DictValue(key, old_dict[key], None))

    for key in common_set:
        output.append(DictValue(key, old_dict[key], new_dict[key]))
        if str(old_dict[key]) != str(new_dict[key]):
            changes += 1

    output.sort(key=attrgetter("key"))
    return changes, output


def format_params_diff(parameter_diff: List[DictValue[Any, Any]]) -> str:
    """Handle the formatting of differences in parameters.

    Args:
        parameter_diff: A list of :class:`DictValue` detailing the differences
            between two dicts returned by :func:`diff_dictionaries`.

    Returns:
        A formatted string that represents a parameter diff

    """
    return (
        "--- Old Parameters\n"
        "+++ New Parameters\n"
        "******************\n"
        + "\n".join(line for v in parameter_diff for line in v.changes())
        + "\n"
    )


def diff_parameters(
    old_params: Dict[str, _OV], new_params: Dict[str, _NV]
) -> List[DictValue[_OV, _NV]]:
    """Compare the old vs. new parameters and returns a "diff".

    If there are no changes, we return an empty list.

    Args:
        old_params: old paramters
        new_params: new parameters

    Returns:
        A list of differences.

    """
    changes, diff = diff_dictionaries(old_params, new_params)
    if changes == 0:
        return []
    return diff


class Action(deploy.Action):
    """Responsible for diffing CloudFormation stacks in AWS and locally.

    Generates the deploy plan based on stack dependencies (these dependencies
    are determined automatically based on references to output values from
    other stacks).

    The plan is then used to create a changeset for a stack using a
    generated template based on the current config.

    """

    DESCRIPTION = "Diff stacks"
    NAME = "diff"

    @property
    def _stack_action(self) -> Callable[..., Status]:
        """Run against a step."""
        return self._diff_stack

    def _diff_stack(self, stack: Stack, **_: Any) -> Status:
        """Handle diffing a stack in CloudFormation vs our config."""
        if self.cancel.wait(0):
            return INTERRUPTED

        if not deploy.should_submit(stack):
            return NotSubmittedStatus()

        provider = self.build_provider()

        if not deploy.should_update(stack):
            stack.set_outputs(provider.get_outputs(stack.fqn))
            return NotUpdatedStatus()

        tags = deploy.build_stack_tags(stack)

        try:
            provider_stack = provider.get_stack(stack.fqn)
        except exceptions.StackDoesNotExist:
            provider_stack = None

        try:
            stack.resolve(self.context, provider)
            parameters = self.build_parameters(stack, provider_stack)
            outputs = provider.get_stack_changes(
                stack, self._template(stack.blueprint), parameters, tags
            )
            stack.set_outputs(outputs)
        except exceptions.StackDidNotChange:
            LOGGER.info("%s:no changes", stack.fqn)
            stack.set_outputs(provider.get_outputs(stack.fqn))
        except exceptions.StackDoesNotExist:
            if self.context.persistent_graph:
                return SkippedStatus(
                    "persistent graph: stack does not exist, will be removed"
                )
            return DoesNotExistInCloudFormation()
        except AttributeError as err:
            if (
                self.context.persistent_graph
                and "defined class or template path" in str(err)
            ):
                return SkippedStatus("persistent graph: will be destroyed")
            raise
        except ClientError as err:
            if (
                err.response["Error"]["Code"] == "ValidationError"
                and "length less than or equal to" in err.response["Error"]["Message"]
            ):
                LOGGER.error(
                    "%s:template is too large to provide directly to the API; "
                    "S3 must be used",
                    stack.name,
                )
                return SkippedStatus("cfngin_bucket: existing bucket required")
            raise
        return COMPLETE

    def run(
        self,
        *,
        concurrency: int = 0,
        dump: Union[bool, str] = False,  # pylint: disable=unused-argument
        force: bool = False,  # pylint: disable=unused-argument
        outline: bool = False,  # pylint: disable=unused-argument
        tail: bool = False,  # pylint: disable=unused-argument
        upload_disabled: bool = False,  # pylint: disable=unused-argument
        **_kwargs: Any,
    ) -> None:
        """Kicks off the diffing of the stacks in the stack_definitions."""
        plan = self._generate_plan(
            require_unlocked=False, include_persistent_graph=True
        )
        plan.outline(logging.DEBUG)
        if plan.keys():
            LOGGER.info("diffing stacks: %s", ", ".join(plan.keys()))
        else:
            LOGGER.warning("no stacks detected (error in config?)")
        walker = build_walker(concurrency)
        plan.execute(walker)

    def pre_run(
        self,
        *,
        dump: Union[bool, str] = False,  # pylint: disable=unused-argument
        outline: bool = False,  # pylint: disable=unused-argument
        **__kwargs: Any,
    ) -> None:
        """Any steps that need to be taken prior to running the action.

        Handle CFNgin bucket access denied & not existing.

        """
        if not self.bucket_name:
            return
        bucket = Bucket(self.context, self.bucket_name, self.bucket_region)
        if bucket.forbidden:
            LOGGER.error("access denied for CFNgin bucket: %s", bucket.name)
            sys.exit(1)
        if bucket.not_found:
            LOGGER.warning(
                'cfngin_bucket "%s" does not exist and will be creating '
                "during the next deploy",
                bucket.name,
            )
            LOGGER.verbose("proceeding without a cfngin_bucket...")
            self.bucket_name = None

    def post_run(
        self, *, dump: Union[bool, str] = False, outline: bool = False, **__kwargs: Any
    ) -> None:
        """Do nothing."""



"""CFNgin environment file parsing."""
from typing import Any, Dict


def parse_environment(raw_environment: str) -> Dict[str, Any]:
    """Parse environment file contents.

    Args:
        raw_environment: Environment file read into a string.

    """
    environment: Dict[str, Any] = {}
    for line in raw_environment.split("\n"):
        line = line.strip()
        if not line:
            continue

        if line.startswith("#"):
            continue

        try:
            key, value = line.split(":", 1)
        except ValueError:
            raise ValueError("Environment must be in key: value format") from None

        environment[key] = value.strip()
    return environment



#!/usr/bin/env python
"""Module with k8s cluster resources."""
import awacs.iam
from awacs.aws import Allow, Condition, PolicyDocument, Statement, StringLike
from awacs.helpers.trust import make_simple_assume_policy
from troposphere import Export, Join, Output, Sub, ec2, eks, iam

from runway.cfngin.blueprints.base import Blueprint
from runway.cfngin.blueprints.variables.types import (
    CFNString,
    EC2SubnetIdList,
    EC2VPCId,
)

IAM_POLICY_ARN_PREFIX = "arn:aws:iam::aws:policy/"


class Cluster(Blueprint):
    """CFNgin blueprint for creating k8s cluster resources."""

    VARIABLES = {
        "EksClusterName": {
            "type": CFNString,
            "description": "Name of the Kubernetes cluster",
            "min_length": 2,
            "max_length": 40,
        },
        "EksSubnets": {
            "type": EC2SubnetIdList,
            "description": "Subnets where the Kubernetes cluster will live.",
        },
        "EksVersion": {"type": CFNString, "description": "Kubernetes version"},
        "VPC": {
            "type": EC2VPCId,
            "description": "VPC where the Kubernetes cluster will live.",
        },
    }

    def create_template(self) -> None:
        """Create template (main function called by CFNgin)."""
        template = self.template
        template.set_version("2010-09-09")
        template.set_description("Kubernetes Master via EKS - V1.0.0")

        # Resources
        ccpsecuritygroup = template.add_resource(
            ec2.SecurityGroup(
                "ClusterControlPlaneSecurityGroup",
                GroupDescription="Cluster communication with worker nodes",
                Tags=[
                    {
                        "Key": Sub("kubernetes.io/cluster/${EksClusterName}"),
                        "Value": "owned",
                    },
                    {"Key": "Product", "Value": "Kubernetes"},
                    {"Key": "Project", "Value": "eks"},
                    {"Key": "Name", "Value": Sub("${EksClusterName}-sg-worker-nodes")},
                ],
                VpcId=self.variables["VPC"].ref,
            )
        )
        template.add_output(
            Output(
                ccpsecuritygroup.title,
                Description="Cluster communication with worker nodes",
                Export=Export(Sub("${AWS::StackName}-ControlPlaneSecurityGroup")),
                Value=ccpsecuritygroup.ref(),
            )
        )

        eksservicerole = template.add_resource(
            iam.Role(
                "EksServiceRole",
                AssumeRolePolicyDocument=make_simple_assume_policy("eks.amazonaws.com"),
                ManagedPolicyArns=[IAM_POLICY_ARN_PREFIX + "AmazonEKSClusterPolicy"],
                Policies=[
                    iam.Policy(
                        PolicyName="EksServiceRolePolicy",
                        PolicyDocument=PolicyDocument(
                            Statement=[
                                Statement(
                                    Action=[
                                        awacs.iam.CreateServiceLinkedRole,
                                        awacs.iam.PutRolePolicy,
                                    ],
                                    Condition=Condition(
                                        StringLike(
                                            "iam:AWSServiceName",
                                            "elasticloadbalancing.amazonaws.com",
                                        )
                                    ),
                                    Effect=Allow,
                                    Resource=[
                                        Sub(
                                            "arn:aws:iam::${AWS::AccountId}:role/"
                                            "aws-service-role/"
                                            "elasticloadbalancing.amazonaws.com/"
                                            "AWSServiceRoleForElasticLoadBalancing*"
                                        )
                                    ],
                                )
                            ]
                        ),
                    )
                ],
            )
        )

        ekscluster = template.add_resource(
            eks.Cluster(
                "EksCluster",
                Name=self.variables["EksClusterName"].ref,
                Version=self.variables["EksVersion"].ref,
                RoleArn=eksservicerole.get_att("Arn"),
                ResourcesVpcConfig=eks.ResourcesVpcConfig(
                    SecurityGroupIds=[ccpsecuritygroup.ref()],
                    SubnetIds=self.variables["EksSubnets"].ref,
                ),
            )
        )
        template.add_output(
            Output(
                f"{ekscluster.title}Name",
                Description="EKS Cluster Name",
                Export=Export(Sub(f"${{AWS::StackName}}-{ekscluster.title}Name")),
                Value=ekscluster.ref(),
            )
        )
        template.add_output(
            Output(
                f"{ekscluster.title}Endpoint",
                Description="EKS Cluster Endpoint",
                Export=Export(Sub(f"${{AWS::StackName}}-{ekscluster.title}Endpoint")),
                Value=ekscluster.get_att("Endpoint"),
            )
        )

        # Additional Outputs
        template.add_output(
            Output(
                "VpcId",
                Description="EKS Cluster VPC Id",
                Export=Export(Sub("${AWS::StackName}-VpcId")),
                Value=self.variables["VPC"].ref,
            )
        )
        template.add_output(
            Output(
                "Subnets",
                Description="EKS Cluster Subnets",
                Export=Export(Sub("${AWS::StackName}-Subnets")),
                Value=Join(",", self.variables["EksSubnets"].ref),
            )
        )


# Helper section to enable easy blueprint -> template generation
# (just run `python <thisfile>` to output the json)
if __name__ == "__main__":
    from runway.context import CfnginContext

    print(  # noqa: T201
        Cluster("test", CfnginContext(parameters={"namespace": "test"})).to_json()
    )



"""Runway static site Module options."""
from __future__ import annotations

from pathlib import Path
from typing import Any, Dict, List, Optional, cast

from pydantic import Extra, root_validator

from ....config.models.base import ConfigProperty


class RunwayStaticSiteExtraFileDataModel(ConfigProperty):
    """Model for Runway static site Module extra_files option item.

    Attributes:
        content_type: An explicit content type for the file. If not provided,
            will attempt to determine based on the name provided.
        content: Inline content that will be used as the file content.
            This or ``file`` must be provided.
        file: Path to an existing file. The content of this file will be uploaded
            to the static site S3 bucket using the name as the object key.
            This or ``content`` must be provided.
        name: The destination name of the file to create.

    """

    content_type: Optional[str] = None
    content: Any = None
    file: Optional[Path] = None
    name: str

    class Config(ConfigProperty.Config):
        """Model configuration."""

        extra = Extra.forbid
        title = "Runway static site Module extra_files option item."

    @root_validator
    def _autofill_content_type(  # pylint: disable=no-self-argument
        cls, values: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Attempt to fill content_type if not provided."""
        if values.get("content_type"):
            return values
        name = cast(str, values.get("name", ""))
        if name.endswith(".json"):
            values["content_type"] = "application/json"
        elif name.endswith(".yaml") or name.endswith(".yml"):
            values["content_type"] = "text/yaml"
        return values

    @root_validator(pre=True)
    def _validate_content_or_file(  # pylint: disable=no-self-argument
        cls, values: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Validate that content or file is provided."""
        if all(i in values and values[i] for i in ["content", "file"]):
            raise ValueError("only one of content or file can be provided")
        if not any(i in values for i in ["content", "file"]):
            raise ValueError("one of content or file must be provided")
        return values


class RunwayStaticSitePreBuildStepDataModel(ConfigProperty):
    """Model for Runway static site Module pre_build_steps option item.

    Attributes:
        command: The command to run.
        cwd: The working directory for the subprocess running the command.
            If not provided, the current working directory is used.

    """

    command: str
    cwd: Path = Path.cwd()

    class Config(ConfigProperty.Config):
        """Model configuration."""

        extra = Extra.forbid
        title = "Runway static site Module pre_build_steps option item."


class RunwayStaticSiteSourceHashingDirectoryDataModel(ConfigProperty):
    """Model for Runway static site Module source_hashing.directory option item.

    Attributes:
        exclusions: List of gitignore formmated globs to ignore when calculating
            the hash.
        path: Path to files to include in the hash.

    """

    exclusions: List[str] = []
    path: Path

    class Config(ConfigProperty.Config):
        """Model configuration."""

        extra = Extra.forbid
        title = "Runway static site Module source_hashing.directories option item."


class RunwayStaticSiteSourceHashingDataModel(ConfigProperty):
    """Model for Runway static site Module source_hashing option.

    Attributes:
        directories: Explicitly provide the directories to use when calculating
            the hash. If not provided, will default to the root of the module.
        enabled: Enable source hashing. If not enabled, build and upload will
            occur on every deploy.
        parameter: SSM parameter where the hash of each build is stored.

    """

    directories: List[RunwayStaticSiteSourceHashingDirectoryDataModel] = [
        RunwayStaticSiteSourceHashingDirectoryDataModel(path="./")  # type: ignore
    ]
    enabled: bool = True
    parameter: Optional[str] = None

    class Config(ConfigProperty.Config):
        """Model configuration."""

        extra = Extra.forbid
        title = "Runway static site Module source_hashing option."


class RunwayStaticSiteModuleOptionsDataModel(ConfigProperty):
    """Model for Runway static site Module options.

    Attributes:
        build_output: Directory where build output is placed. Defaults to current
            working directory.
        build_steps: List of commands to run to build the static site.
        extra_files: List of files that should be uploaded to S3 after the build.
            Used to dynamically create or select file.
        pre_build_steps: Commands to be run prior to the build process.
        source_hashing: Overrides for source hash calculation and tracking.

    """

    build_output: str = "./"
    build_steps: List[str] = []
    extra_files: List[RunwayStaticSiteExtraFileDataModel] = []
    pre_build_steps: List[RunwayStaticSitePreBuildStepDataModel] = []
    source_hashing: RunwayStaticSiteSourceHashingDataModel = (
        RunwayStaticSiteSourceHashingDataModel()
    )

    class Config(ConfigProperty.Config):
        """Model configuration."""

        extra = Extra.ignore
        title = "Runway static site Module options."



"""Type definitions."""
from __future__ import annotations

from typing import Optional

from typing_extensions import TypedDict


class AwsLambdaHookDeployResponseTypedDict(TypedDict):
    """Dict output of :class:`runway.cfngin.hooks.awslambda.models.response.AwsLambdaHookDeployResponse` using aliases."""  # noqa

    CodeSha256: str
    Runtime: str
    S3Bucket: str
    S3Key: str
    S3ObjectVersion: Optional[str]



"""CFNgin package source models."""
# pylint: disable=no-self-argument
from __future__ import annotations

from typing import Any, Dict, List, Optional

from pydantic import Extra, Field, root_validator

from ..base import ConfigProperty


class GitCfnginPackageSourceDefinitionModel(ConfigProperty):
    """Model for a git package source definition.

    Package source located in a git repository.

    Attributes:
        branch: Branch name.
        commit: Commit hash.
        configs: List of CFNgin config paths to execute.
        paths: List of paths to append to ``sys.path``.
        tag: Git tag.
        uri: Remote git repo URI.

    """

    branch: Optional[str] = Field(
        default=None, title="Git Branch", examples=["ENV-dev", "ENV-prod", "master"]
    )
    commit: Optional[str] = Field(default=None, title="Git Commit Hash")
    configs: List[str] = Field(
        default=[],
        description="Array of paths relative to the root of the package source "
        "for configuration that should be merged into the current configuration file.",
    )
    paths: List[str] = Field(
        default=[],
        description="Array of paths relative to the root of the package source to add to $PATH.",
    )
    tag: Optional[str] = Field(
        default=None, title="Git Tag", examples=["1.0.0", "v1.0.0"]
    )
    uri: str = Field(
        ...,
        title="Git Repository URI",
        examples=["git@github.com:onicagroup/runway.git"],
    )

    class Config(ConfigProperty.Config):
        """Model configuration."""

        extra = Extra.forbid
        schema_extra: Dict[str, Any] = {
            "description": "Information about git repositories that should be included "
            "in the processing of this configuration file."
        }
        title = "CFNgin Git Repository Package Source Definition"

    @root_validator
    def _validate_one_ref(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Ensure that only one ref is defined."""
        ref_keys = ["branch", "commit", "tag"]
        count_ref_defs = sum(bool(values.get(i)) for i in ref_keys)
        if count_ref_defs > 1:
            raise ValueError(f"only one of {ref_keys} can be defined")
        return values


class LocalCfnginPackageSourceDefinitionModel(ConfigProperty):
    """Model for a CFNgin local package source definition.

    Package source located on a local disk.

    Attributes:
        configs: List of CFNgin config paths to execute.
        paths: List of paths to append to ``sys.path``.
        source: Source.

    """

    configs: List[str] = Field(
        default=[],
        description="Array of paths relative to the root of the package source "
        "for configuration that should be merged into the current configuration file.",
    )
    paths: List[str] = Field(
        default=[],
        description="Array of paths relative to the root of the package source to add to $PATH.",
    )
    source: str = Field(
        ...,
        description="Path relative to the current configuration file that is the "
        "root of the local package source.",
    )

    class Config(ConfigProperty.Config):
        """Model configuration."""

        extra = Extra.forbid
        schema_extra: Dict[str, Any] = {
            "description": "Information about local directories that should be "
            "included in the processing of this configuration file."
        }
        title = "CFNgin Local Package Source Definition"


class S3CfnginPackageSourceDefinitionModel(ConfigProperty):
    """Model for a CFNgin S3 package source definition.

    Package source located in AWS S3.

    Attributes:
        bucket: AWS S3 bucket name.
        configs: List of CFNgin config paths to execute.
        key: Object key. The object should be a zip file.
        paths: List of paths to append to ``sys.path``.
        requester_pays: AWS S3 requester pays option.
        use_latest: Use the latest version of the object.

    """

    bucket: str = Field(..., title="AWS S3 Bucket Name")
    configs: List[str] = Field(
        default=[],
        description="Array of paths relative to the root of the package source "
        "for configuration that should be merged into the current configuration file.",
    )
    key: str = Field(..., title="AWS S3 Object Key")
    paths: List[str] = Field(
        default=[],
        description="Array of paths relative to the root of the package source to add to $PATH.",
    )
    requester_pays: bool = Field(
        default=False,
        description="Confirms that the requester knows that they will be charged "
        "for the request.",
    )
    use_latest: bool = Field(
        default=True,
        description="Update the local copy if the last modified date in AWS S3 changes.",
    )

    class Config(ConfigProperty.Config):
        """Model configuration."""

        extra = Extra.forbid
        schema_extra: Dict[str, Any] = {
            "description": "Information about a AWS S3 objects that should be "
            "downloaded, unzipped, and included in the processing of "
            "this configuration file."
        }
        title = "CFNgin S3 Package Source Definition"


class CfnginPackageSourcesDefinitionModel(ConfigProperty):
    """Model for a CFNgin package sources definition.

    Attributes:
        git: Package source located in a git repo.
        local: Package source located on a local disk.
        s3: Package source located in AWS S3.

    """

    git: List[GitCfnginPackageSourceDefinitionModel] = Field(
        default=[],
        title="CFNgin Git Repository Package Source Definitions",
        description=GitCfnginPackageSourceDefinitionModel.Config.schema_extra[
            "description"
        ],
    )
    local: List[LocalCfnginPackageSourceDefinitionModel] = Field(
        default=[],
        title="CFNgin Local Package Source Definitions",
        description=LocalCfnginPackageSourceDefinitionModel.Config.schema_extra[
            "description"
        ],
    )
    s3: List[S3CfnginPackageSourceDefinitionModel] = Field(
        default=[],
        title="CFNgin S3 Package Source Definitions",
        description=S3CfnginPackageSourceDefinitionModel.Config.schema_extra[
            "description"
        ],
    )

    class Config(ConfigProperty.Config):
        """Model configuration."""

        extra = Extra.forbid
        schema_extra: Dict[str, Any] = {
            "description": "Map of additional package sources to include when "
            "processing this configuration file."
        }
        title = "CFNgin Package Sources Definition"



"""Callback URL Retriever.

Dependency pre-hook responsible for ensuring correct
callback urls are retrieved or a temporary one is used in it's place.

"""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Dict, Optional

from ...base import HookArgsBaseModel

if TYPE_CHECKING:
    from .....context import CfnginContext

LOGGER = logging.getLogger(__name__)


class HookArgs(HookArgsBaseModel):
    """Hook arguments."""

    stack_name: str
    """The name of the stack to check against."""

    user_pool_arn: Optional[str] = None
    """The ARN of the User Pool to check for a client."""


def get(context: CfnginContext, *__args: Any, **kwargs: Any) -> Dict[str, Any]:
    """Retrieve the callback URLs for User Pool Client Creation.

    When the User Pool is created a Callback URL is required. During a post
    hook entitled ``client_updater`` these Callback URLs are updated to that
    of the Distribution. Before then we need to ensure that if a Client
    already exists that the URLs for that client are used to prevent any
    interuption of service during deploy.

    Arguments parsed by
    :class:`~runway.cfngin.hooks.staticsite.auth_at_edge.callback_url_retriever.HookArgs`.

    Args:
        context: The context instance.

    """
    args = HookArgs.parse_obj(kwargs)
    session = context.get_session()
    cloudformation_client = session.client("cloudformation")
    cognito_client = session.client("cognito-idp")

    context_dict = {"callback_urls": ["https://example.org"]}
    try:
        # Return the current stack if one exists
        stack_desc = cloudformation_client.describe_stacks(StackName=args.stack_name)
        # Get the client_id from the outputs
        outputs = stack_desc["Stacks"][0]["Outputs"]

        if args.user_pool_arn:
            user_pool_id = args.user_pool_arn.split("/")[-1:][0]
        else:
            user_pool_id = [
                o["OutputValue"]
                for o in outputs
                if o["OutputKey"] == "AuthAtEdgeUserPoolId"
            ][0]

        client_id = [
            o["OutputValue"] for o in outputs if o["OutputKey"] == "AuthAtEdgeClient"
        ][0]

        # Poll the user pool client information
        resp = cognito_client.describe_user_pool_client(
            UserPoolId=user_pool_id, ClientId=client_id
        )

        # Retrieve the callbacks
        callbacks = resp["UserPoolClient"]["CallbackURLs"]

        if callbacks:
            context_dict["callback_urls"] = callbacks
        return context_dict
    except Exception:  # pylint: disable=broad-except
        return context_dict



"""Terraform version management."""
from __future__ import annotations

import hashlib
import json
import locale
import logging
import os
import platform
import re
import shutil
import subprocess
import sys
import tempfile
import zipfile
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generator,
    List,
    Optional,
    Union,
    cast,
    overload,
)
from urllib.error import URLError
from urllib.request import urlretrieve

import hcl
import hcl2
import requests
from packaging.version import InvalidVersion
from typing_extensions import Final

from ..compat import cached_property
from ..exceptions import HclParserError
from ..utils import FileHash, Version, get_hash_for_filename, merge_dicts
from . import EnvManager, handle_bin_download_error

if TYPE_CHECKING:
    from pathlib import Path
    from types import ModuleType

    from .._logging import RunwayLogger

LOGGER = cast("RunwayLogger", logging.getLogger(__name__))
TF_VERSION_FILENAME = ".terraform-version"


def download_tf_release(
    version: str,
    versions_dir: Path,
    command_suffix: str,
    tf_platform: Optional[str] = None,
    arch: Optional[str] = None,
) -> None:
    """Download Terraform archive and return path to it."""
    version_dir = versions_dir / version

    if arch is None:
        arch = os.getenv("TFENV_ARCH", "amd64")

    if tf_platform:
        tfver_os = tf_platform + "_" + arch
    else:
        if platform.system().startswith("Darwin"):
            tfver_os = f"darwin_{arch}"
        elif platform.system().startswith("Windows") or (
            platform.system().startswith("MINGW64")
            or (
                platform.system().startswith("MSYS_NT")
                or platform.system().startswith("CYGWIN_NT")
            )
        ):
            tfver_os = f"windows_{arch}"
        else:
            tfver_os = f"linux_{arch}"

    download_dir = tempfile.mkdtemp()
    filename = f"terraform_{version}_{tfver_os}.zip"
    shasums_name = f"terraform_{version}_SHA256SUMS"
    tf_url = "https://releases.hashicorp.com/terraform/" + version

    try:
        LOGGER.verbose("downloading Terraform from %s...", tf_url)
        for i in [filename, shasums_name]:
            urlretrieve(tf_url + "/" + i, os.path.join(download_dir, i))
    except URLError as exc:
        handle_bin_download_error(exc, "Terraform")

    tf_hash = get_hash_for_filename(filename, os.path.join(download_dir, shasums_name))
    checksum = FileHash(hashlib.sha256())
    checksum.add_file(os.path.join(download_dir, filename))
    if tf_hash != checksum.hexdigest:
        LOGGER.error(
            "downloaded Terraform %s does not match sha256 %s", filename, tf_hash
        )
        sys.exit(1)

    with zipfile.ZipFile(os.path.join(download_dir, filename)) as tf_zipfile:
        version_dir.mkdir(parents=True, exist_ok=True)
        tf_zipfile.extractall(str(version_dir))

    shutil.rmtree(download_dir)
    result = version_dir / ("terraform" + command_suffix)
    result.chmod(result.stat().st_mode | 0o0111)  # ensure it is executable


def get_available_tf_versions(include_prerelease: bool = False) -> List[str]:
    """Return available Terraform versions."""
    tf_releases = json.loads(
        requests.get("https://releases.hashicorp.com/index.json").text
    )["terraform"]

    # Remove versions that don't align with
    # PEP440 (https://peps.python.org/pep-0440/)
    for k, _v in list(tf_releases["versions"].items()):
        try:
            Version(k)
        except InvalidVersion:
            LOGGER.debug("InvalidVersion found, skipping %s. ", k)
            del tf_releases["versions"][k]

    tf_versions = sorted(
        [k for k, _v in tf_releases["versions"].items()],  # descending
        key=Version,
        reverse=True,
    )

    if include_prerelease:
        return [i for i in tf_versions if i]
    return [i for i in tf_versions if i and "-" not in i]


def get_latest_tf_version(include_prerelease: bool = False) -> str:
    """Return latest Terraform version."""
    return get_available_tf_versions(include_prerelease)[0]


def load_terraform_module(parser: ModuleType, path: Path) -> Dict[str, Any]:
    """Load all Terraform files in a module into one dict.

    Args:
        parser (Union[hcl, hcl2]): Parser to use when loading files.
        path: Terraform module path. All Terraform files in the
            path will be loaded.

    """
    result: Dict[str, Any] = {}
    LOGGER.debug("using %s parser to load module: %s", parser.__name__.upper(), path)
    for tf_file in path.glob("*.tf"):
        try:
            tf_config = parser.loads(tf_file.read_text())  # type: ignore
            result = merge_dicts(result, cast(Dict[str, Any], tf_config))
        except Exception as exc:
            raise HclParserError(exc, tf_file, parser) from None
    return result


class TFEnvManager(EnvManager):
    """Terraform version management.

    Designed to be compatible with https://github.com/tfutils/tfenv.

    """

    VERSION_REGEX: Final[str] = r"^(Terraform v)?(?P<version>[0-9]+\.[0-9]+\.[0-9]+\S*)"
    VERSION_OUTPUT_REGEX: Final[
        str
    ] = r"^Terraform v(?P<version>[0-9]*\.[0-9]*\.[0-9]*)(?P<suffix>-.*)?"

    def __init__(self, path: Optional[Path] = None) -> None:
        """Initialize class."""
        super().__init__("terraform", "tfenv", path)

    @cached_property
    def backend(self) -> Dict[str, Any]:
        """Backend config of the Terraform module."""
        # Terraform can only have one backend configured; this formats the
        # data to make it easier to work with
        return [
            {"type": k, "config": v}
            for k, v in self.terraform_block.get(
                "backend", {None: cast(Dict[str, str], {})}
            ).items()
        ][0]

    @cached_property
    def terraform_block(self) -> Dict[str, Any]:
        """Collect Terraform configuration blocks from a Terraform module."""

        @overload
        def _flatten_lists(data: Dict[str, Any]) -> Dict[str, Any]:
            ...

        @overload
        def _flatten_lists(data: List[Any]) -> List[Any]:
            ...

        @overload
        def _flatten_lists(data: str) -> str:
            ...

        def _flatten_lists(
            data: Union[Dict[str, Any], List[Any], Any]
        ) -> Union[Dict[str, Any], Any]:
            """Flatten HCL2 list attributes until its fixed.

            python-hcl2 incorrectly turns all attributes into lists so we need
            to flatten them so they are more similar to HCL.

            https://github.com/amplify-education/python-hcl2/issues/6

            Args:
                data: Dict with lists to flatten.

            """
            if not isinstance(data, dict):
                return data
            copy_data = cast(Dict[str, Any], data.copy())
            for attr, val in copy_data.items():
                if isinstance(val, list):
                    if len(cast(List[Any], val)) == 1:
                        # pull single values out of lists
                        data[attr] = _flatten_lists(cast(Any, val[0]))
                    else:
                        data[attr] = [_flatten_lists(v) for v in cast(List[Any], val)]
                elif isinstance(val, dict):
                    data[attr] = _flatten_lists(cast(Dict[str, Any], val))
            return data

        try:
            result: Union[Dict[str, Any], List[Dict[str, Any]]] = load_terraform_module(
                hcl2, self.path
            ).get("terraform", cast(Dict[str, Any], {}))
        except HclParserError as exc:
            LOGGER.warning(exc)
            LOGGER.warning("failed to parse as HCL2; trying HCL...")
            try:
                result = load_terraform_module(hcl, self.path).get(
                    "terraform", cast(Dict[str, Any], {})
                )
            except HclParserError as exc2:
                LOGGER.warning(exc2)
                # return an empty dict if we can't parse HCL
                # let Terraform decide if it's actually valid
                result = {}

        # python-hcl2 turns all blocks into lists in v0.3.0. this flattens it.
        if isinstance(result, list):
            return _flatten_lists({k: v for i in result for k, v in i.items()})
        return _flatten_lists(result)

    @cached_property
    def version(self) -> Optional[Version]:
        """Terraform version."""
        version_requested = self.current_version or self.get_version_from_file()

        if not version_requested:
            return None

        if re.match(r"^min-required$", version_requested):
            LOGGER.debug("tfenv: detecting minimal required version")
            version_requested = self.get_min_required()

        if re.match(r"^latest:.*$", version_requested):
            regex = re.search(r"latest:(.*)", version_requested).group(  # type: ignore
                1
            )
            include_prerelease_versions = False
        elif re.match(r"^latest$", version_requested):
            regex = r"^[0-9]+\.[0-9]+\.[0-9]+$"
            include_prerelease_versions = False
        else:
            regex = f"^{version_requested}$"
            include_prerelease_versions = True
            # Return early (i.e before reaching out to the internet) if the
            # matching version is already installed
            if (self.versions_dir / version_requested).is_dir():
                self.current_version = version_requested
                return self.parse_version_string(self.current_version)

        try:
            version = next(
                i
                for i in get_available_tf_versions(include_prerelease_versions)
                if re.match(regex, i)
            )
        except StopIteration:
            LOGGER.error("unable to find a Terraform version matching regex: %s", regex)
            sys.exit(1)
        self.current_version = version
        return self.parse_version_string(self.current_version)

    @cached_property
    def version_file(self) -> Optional[Path]:
        """Find and return a ".terraform-version" file if one is present.

        Returns:
            Path to the Terraform version file.

        """
        for path in [self.path, self.path.parent]:
            test_path = path / TF_VERSION_FILENAME
            if test_path.is_file():
                LOGGER.debug("using version file: %s", test_path)
                return test_path
        return None

    def get_min_required(self) -> str:
        """Get the defined minimum required version of Terraform.

        Returns:
            The minimum required version as defined in the module.

        """
        version = self.terraform_block.get("required_version")

        if version:
            if re.match(r"^!=.+", version):
                LOGGER.error(
                    "min required Terraform version is a negation (%s) "
                    "- unable to determine required version",
                    version,
                )
                sys.exit(1)
            else:
                version = re.search(r"[0-9]*\.[0-9]*(?:\.[0-9]*)?", version)
                if version:
                    LOGGER.debug("detected minimum Terraform version is %s", version)
                    return version.group(0)
        LOGGER.error(
            "Terraform version specified as min-required, but unable to "
            "find a specified version requirement in this module's tf files"
        )
        sys.exit(1)

    def get_version_from_file(self, file_path: Optional[Path] = None) -> Optional[str]:
        """Get Terraform version from a file.

        Args:
            file_path: Path to file that will be read.

        """
        file_path = file_path or self.version_file
        if file_path and file_path.is_file():
            return file_path.read_text(
                encoding=locale.getpreferredencoding(do_setlocale=False)
            ).strip()
        LOGGER.debug("file path not provided and version file could not be found")
        return None

    def install(self, version_requested: Optional[str] = None) -> str:
        """Ensure Terraform is available."""
        if version_requested:
            self.set_version(version_requested)

        if not self.version:
            raise ValueError(
                f"version not provided and unable to find a {TF_VERSION_FILENAME} file"
            )

        # Now that a version has been selected, skip downloading if it's
        # already been downloaded
        if (self.versions_dir / str(self.version)).is_dir():
            LOGGER.verbose(
                "Terraform version %s already installed; using it...", self.version
            )
            return str(self.bin)

        LOGGER.info("downloading and using Terraform version %s ...", self.version)
        download_tf_release(str(self.version), self.versions_dir, self.command_suffix)
        LOGGER.verbose("downloaded Terraform %s successfully", self.version)
        return str(self.bin)

    def list_installed(self) -> Generator[Path, None, None]:
        """List installed versions of Terraform.

        Only lists versions of Terraform that have been installed by an instance
        if this class or by tfenv.

        """
        LOGGER.verbose("checking %s for Terraform versions...", self.versions_dir)
        return self.versions_dir.rglob("*.*.*")

    def set_version(self, version: str) -> None:
        """Set current version.

        Clears cached values as needed.

        Args:
            version: Version string. Must be in the format of
                ``<major>.<minor>.<patch>`` with an optional ``-<prerelease>``.

        """
        if self.current_version == version:
            return
        self.current_version = version
        self._del_cached_property("version")

    @classmethod
    def get_version_from_executable(
        cls,
        bin_path: Union[Path, str],
        *,
        cwd: Optional[Union[Path, str]] = None,
        env: Optional[Dict[str, str]] = None,
    ) -> Optional[Version]:
        """Get Terraform version from an executable.

        Args:
            bin_path: Path to the Terraform binary to retrieve the version from.
            cwd: Current working directory to use when calling the executable.
            env: Environment variable overrides.

        """
        output = subprocess.check_output(
            [str(bin_path), "-version"], cwd=cwd, env=env
        ).decode()
        match = re.search(cls.VERSION_OUTPUT_REGEX, output)
        if not match:
            return None
        return cls.parse_version_string(output)

    @classmethod
    def parse_version_string(cls, version: str) -> Version:
        """Parse version string into a :class:`Version`.

        Args:
            version: Version string to parse. Must be in the format of
                ``<major>.<minor>.<patch>`` with an optional ``-<prerelease>``.

        """
        match = re.search(cls.VERSION_REGEX, version)
        if not match:
            raise ValueError(
                f"provided version doesn't conform to regex: {cls.VERSION_REGEX}"
            )
        return Version(match.group("version"))
